/*
 *
 *
 *    CwModem.ino
 *
 *    A modem for Morse CW (A1A).
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

//
//  NOTES:
//
//    1. This firmware WILL write to the EEPROM area of the processor.  It
//       will also attempt to wear-level the EEPROM to prevent premature wear
//       of the EEPROM.
//    2. If upgrading from a previous version, you should execute the command
//       #EC=1; from the serial port to clear the previous EEPROM contents.
//    3. Several configuration options follow, and are marked either REQUIRED,
//       OPTIONAL, or ADVANCED.  The values provided are reasonable defaults,
//       and you should avoid changing any configuration options unless you
//       are sure of the changes that will result.
//

#include "targets.h"

//
//  REQUIRED: Select the CW recevier data input source.
//
//  This setting will cause another header to be included, which contains the
//  settings specific to the input method selected.  Once you select one of the
//  available options, make sure to look at the corresponding header file, to
//  review the options for that input method.
//
//  Available values for CW_DATA_SOURCE:
//
//      CW_SOURCE_567 - to use an LM567 or equivalent detector circuit; edit
//            the options in the file "config_567.h"; this is the default
//            input method.
//      CW_SOURCE_AA0 - to read audio data directly from the A0 pin on an
//            Arduino board, as described here:
//                   http://www.kk5jy.net/cw-modem-v1/
//            Edit the options in the file "config_AA0.h"
//      CW_SOURCE_TSH - to read audio data directly from the line input of
//            the Teensy audio adaptor board:
//                   https://www.pjrc.com/store/teensy3_audio.html
//            Edit the options in the file "config_TSH.h"
//
#define CW_DATA_SOURCE CW_SOURCE_TSH

//
//  REQUIRED: CW TX (key) output (digital output)
//
//  This is the pin used to transmit key pulses. This needs to be connected
//  to some kind of output switching device, like an n-channel FET, NPN BJT,
//  or solid state relay, which is then connected to the KEY input on the
//  transmitter.  This pin will be driven HIGH to generate key closures.
//
//  This pin can also be attached to a current-limited LED as a transmit
//  key-down indicator.
//
#define CW_KEY_PIN (2)

//
//  OPTIONAL: PTT output (digital output)
//
//  This is the pin used to enable the radio transmitter.  If used, this
//  needs to be connected to some kind of output switching device, like an
//  n-channel FET, NPN BJT, or solid state relay, which is then connected
//  to the PTT input on the transmitter.  This pin will be driven HIGH to
//  enable the transmitter.
//
//  This pin can also be attached to a current-limited LED as a transmit
//  PTT indicator.
//
//  Set this value to zero (0) to disable PTT output.
//
#define CW_PTT_PIN (3)

//
//  OPTIONAL: Command LED (digital output)
//
//  If attached to an LED, the LED will blink momentarily when data is
//  received from the host.
//
//  Set this value to zero (0) to disable.
//
//  This is disabled by default, since most boards have LEDs attached to
//  the serial port.  Set to something other than zero to drive a custom
//  LED attached to another port.
//
#define CMD_LED (0)

//
//  OPTIONAL: CW detector LED
//
//  If attached to an LED, the LED will blink whenever tone is detected,
//  either by the external decoder IC, or by the software detector.
//
//  Leave this set to 13 to blink the built-in LED attached to D13 on most
//  Arduinio boards.  Set this to something else on Teensy, since D13 is
//  also used for the SCK signal to the audio shield.
//
//  Set this value to zero (0) to disable.
//
#define CW_LED_PIN (4)

//
//  REQUIRED: Host Serial Port
//
//  Assign a specific hardware port to be the interface to the host;
//  by default, the USB connection on most boards is 'Serial', which
//  is also the hardware TTL serial port on the UNO board.  The hardware
//  serial device on the Leonardo is Serial1.
//
//  Leave this alone unless you have a Leonardo, or want to use a
//  non-default host port on a Mega board.
//
#define HostSerial Serial

//
//  REQUIRED: host port serial rate
//
//  Set the 'baud rate' of the serial port, which is the speed you will
//  need to use to communicate with the modem from the host system.
//
#define SerialRate (9600)

// ================== END OF CONFIGURATION SETTINGS ==================
//
//  ...everything below this point is the sketch code; do not change
//     anything below this unless you know what you are doing...
//

// input headers
#if CW_DATA_SOURCE == CW_SOURCE_567
	#include "config_567.h"
#elif CW_DATA_SOURCE == CW_SOURCE_AA0
	#include "config_AA0.h"
#elif CW_DATA_SOURCE == CW_SOURCE_TSH
	#include "config_TSH.h"
#endif

// local headers
#include "CircularBuffer.h"
#include "CwTimingLogic.h"
#include "CwDecoderLogic.h"
#include "SplitCommand.h"
#include "Elapsed.h"
#include "Bounce2.h"
#include "eeblock.h"

// headers needed to read analog audio from A0 pin
#if CW_DATA_SOURCE == CW_SOURCE_AA0 || CW_DATA_SOURCE == CW_SOURCE_TSH
	#include "es.h"

	#if defined(CW_TRIGRAPH_LEFT) && defined(CW_TRIGRAPH_CENTER) && defined(CW_TRIGRAPH_RIGHT)
	#define CW_TRIGRAPH
	#include "TriGraph.h"
	#endif

	#if CW_DATA_SOURCE == CW_SOURCE_AA0
		#include "avraudio2.h"

	#elif CW_DATA_SOURCE == CW_SOURCE_TSH
		#if CW_RATE_DIVISOR > 1 && CW_ORDER == 0
		#error Use of downsampling requires the preselector order of >= 1
		#endif

		// Fs is fixed on the audio adaptor
		#define CW_SAMPLE_RATE (44100 / CW_RATE_DIVISOR)

		// this is the packet size returned by the codec
		#define SG_PACKET_SIZE (128)

		// include Teensy headers for audio
		#include <Audio.h>
		#include <Wire.h>

		//
		//  Teensy Audio Library Configuration
		//
		//  NOTE: this configuration reads the LEFT channel only
		//        from the audio shield, to obtain mono audio.
		//

		// GUItool: begin automatically generated code
		AudioControlSGTL5000     sgtl5000;
		AudioInputI2S            i2sIn;
		AudioFilterBiquad        iirIn;
		AudioRecordQueue         inQueue;
		AudioAnalyzePeak         peakIn;
		AudioConnection          patchCord1(i2sIn, 0, iirIn, 0);
		AudioConnection          patchCord2(iirIn, inQueue);
		AudioConnection          patchCord3(iirIn, peakIn);
		// GUItool: end automatically generated code
	#endif

	// headers needed for decoders
	#if EMULATE_LM567 == true
		// include 567 emulation class
		#include "em567.h"
	#else
		// include delay-line class
		#include "tone.h"
	#endif
#endif

// the version string
#define VERSION_STRING ("CwModem 2.0 (beta-20171108a)")

// CR+LF
#define LOCAL_CRLF ("\r\n")

// sequence to send to backspace one char and delete it
#define BACKSPACE_STRING ("\x08 \x08")

// import some namespaces
using namespace KK5JY::Collections;
using namespace KK5JY::CW;

// buffer for element timing data
static CircularBuffer<CwElement> CwBuffer(32);

// buffer for decoded elements
static CircularBuffer<DataElement> ElementBuffer(32);

// a TX buffer
#if CW_DATA_SOURCE == CW_SOURCE_TSH
static CircularBuffer<char> TxBuffer(64);
#else
static CircularBuffer<char> TxBuffer(32);
#endif

// clock for CW
static CwTimingLogic CwTiming;

// decoder for CW
static CwDecoderLogic CwDecoder;

// the clock restoration logic
static TimingLogic *Timing = &CwTiming;

// the decoder
static DecoderLogic *Decoder = &CwDecoder;

#if CW_DATA_SOURCE == CW_SOURCE_AA0 || CW_DATA_SOURCE == CW_SOURCE_TSH
	// RX CW state - start with LOW (unkeyed or no-tone)
	static bool lastState = LOW;

	// audio detector output state; emulates a 567 output, so key-down is LOW
	static volatile bool audioState = HIGH;

	#if CW_DATA_SOURCE == CW_SOURCE_TSH
		static unsigned long audioClock = 0;
		static uint8_t audioClockCtr = 0;
		static uint8_t audioClockLeap = 0;
		static uint8_t audioClockTarget = 44;
		#if CW_RATE_DIVISOR != 1
			static uint8_t downsample_ctr = 0;
		#endif
	#elif CW_DATA_SOURCE == CW_SOURCE_AA0
		// stores the audio level for diagnostics
		static volatile uint8_t audio_max = 0;
	#endif

	#if EMULATE_LM567 == true
		// define the detector type to be LM567 emulator
		#if CW_DATA_SOURCE == CW_SOURCE_TSH
			typedef float vco_counter_t;
		#else
			typedef uint8_t vco_counter_t;
		#endif

		#define \
			Detector \
			KK5JY::DSP::Tone567<int16_t, float>

		// allocate the detector object
		static Detector detector(Detector::Config::Create(
			CW_FREQ,
			CW_SAMPLE_RATE,
			EM567_OUTPUT_INTEGRATOR,
			EM567_LOOP_INTEGRATOR,
			EM567_LOOP_VCO_LIMIT));

		#if CW_DATA_SOURCE == CW_SOURCE_TSH
			typedef float smoother_t;

			// convert constants to user settings
			static int32_t pllTrackOffset = EM567_TRACK_OFFSET;
			static int32_t pllTrackGain = EM567_TRACK_GAIN;
			static smoother_t pllTrackAlpha = EM567_TRACK_ALPHA;
			static smoother_t pllTrackLowScale = EM567_TRACK_LOW_SCALE;

			// low-pass filter for PLL deviation value; used for a tuning indicator
			static KK5JY::DSP::Smoother<smoother_t> pllOffsetLPF(pllTrackAlpha, -pllTrackOffset);

			// storage for the output of the LPF (scaled and shifted)
			static int16_t pllOffset = 0;

			// timestamp of the last PLL offset tracking update; used for latching
			static unsigned long lastPllOffsetUpdate = 0;

			// flag indicating that PLL offset is actively being reported; used for latching
			static bool pllOffsetActive = false;

		#endif

		#ifdef CW_TRIGRAPH
			static uint8_t triCtr = 0;
			const uint8_t triTarget = 1; // process at least this many frames between LED updates
		#endif

	#else
		// define the detector type to be delay-line-product
		#define \
			Detector \
			KK5JY::DSP::ToneDetector<int16_t, KK5JY::DSP::FixedAverager<int32_t, int32_t>, KK5JY::DSP::FixedPoint<int16_t> >

		// allocate the detector object
		static Detector detector(Detector::Config::Create(
			CW_FREQ,
			CW_SAMPLE_RATE));

	#endif
#elif CW_DATA_SOURCE == CW_SOURCE_567
	// debounce for the key input digital pin
	static Bounce KeyInput;

	// RX CW state - start with 'true' (unkeyed or no-tone)
	static bool lastState = true;
#endif

#ifdef CW_TRIGRAPH
	KK5JY::TriGraph trigraph(CW_TRIGRAPH_LEFT, CW_TRIGRAPH_CENTER, CW_TRIGRAPH_RIGHT);
#endif

// a string to hold incoming data
static String inputString; // start with Version, WPM queries

// and another for command data
static String cmdText = "";

// command parsing
static String command = "";
static String argument = "";

// whether the string is complete
static boolean inputReady = false;

// I/O buffer
static char ioBuffer[32];

#if CMD_LED > 0
	// data LED hang-time (makes it more obvious to user that data was rec'd)
	#define DataLedDelay (250)
	// last time data was rec'd from the host
	static unsigned long lastRxData = 0;
	// cmd LED state
	static uint8_t lastCmdLED = LOW;
#endif

// last CW transition
static unsigned long lastTime = 0;

// used to detect end of transmission
static bool hanging = false;

// echo USB RX data back to host (0 = no, 1 = yes, 2 = data only)
static uint8_t echo = false;

// currently parsing a command
static bool cmd = false;

// TX mode (otherwise RX)
static bool tx = false;

// tracks PTT state without reading PTT pin
static bool pttState = false;

// TX element timing
static unsigned long curTxStart = 0;
static unsigned long lastKeyDown = 0;
static unsigned curTxLength = 0;
static CwElement cwOut;

// settings block (will be persisted)
struct SettingsBlock {
	// PTT lead-in (msec)
	unsigned pttLeadIn;
	// PTT lead-out (msec)
	unsigned pttLeadOut;
	// true if CR and/or LF should be converted to BT when sending
	bool convertCRLF;

	// CwTiming settings
	uint16_t boxcarLength;
	float minimumAverageDistance;
	SpeedSources txMode, rxMode;
	uint8_t txSpeed, rxSpeed;

	#if CW_DATA_SOURCE == CW_SOURCE_567
	// digital input debounce interval
	uint16_t inputBounce;
	#endif

	#if CW_DATA_SOURCE == CW_SOURCE_TSH
	// PLL settings
	uint16_t I;
	uint16_t Q;
	uint16_t maxCV;

	// PLL reporting scale, offset, and correction
	int16_t pllReportingGain;
	int16_t pllReportingOffset;
	float pllLowCorrection;

	// line-input level
	uint8_t line_input;
	#endif

	// default ctor - assemble some reasonable default values from different places
	SettingsBlock() {
		pttLeadIn = 100;
		pttLeadOut = 250;
		convertCRLF = false;

		boxcarLength = DefaultBoxCarLength;
		minimumAverageDistance = DefaultMinimumAverageDistance;
		txMode = SpeedAuto;
		txSpeed = 25;
		rxMode = SpeedAuto;
		rxSpeed = 25;

		#if CW_DATA_SOURCE == CW_SOURCE_567
		inputBounce = DefaultKeyDebounceInterval;
		#endif

		#if CW_DATA_SOURCE == CW_SOURCE_TSH
		I = EM567_OUTPUT_INTEGRATOR;
		Q = EM567_LOOP_INTEGRATOR;
		maxCV = EM567_LOOP_VCO_LIMIT;

		pllReportingGain = EM567_TRACK_GAIN;
		pllReportingOffset = EM567_TRACK_OFFSET;
		pllLowCorrection = EM567_TRACK_LOW_SCALE;

		line_input = CW_INPUT_LEVEL;
		#endif
	}

	// copy ctor
	SettingsBlock(const SettingsBlock &sb) {
		pttLeadIn = sb.pttLeadIn;
		pttLeadOut = sb.pttLeadOut;
		convertCRLF = sb.convertCRLF;

		boxcarLength = sb.boxcarLength;
		minimumAverageDistance = sb.minimumAverageDistance;
		txMode = sb.txMode;
		txSpeed = sb.txSpeed;
		rxMode = sb.rxMode;
		rxSpeed = sb.txSpeed;

		#if CW_DATA_SOURCE == CW_SOURCE_TSH
		I = sb.I;
		Q = sb.Q;
		maxCV = sb.maxCV;

		pllReportingGain = sb.pllReportingGain;
		pllReportingOffset = sb.pllReportingOffset;
		pllLowCorrection = sb.pllLowCorrection;

		line_input = sb.line_input;
		#endif

		#if CW_DATA_SOURCE == CW_SOURCE_567
		inputBounce = sb.inputBounce;
		#endif
	}

	// operator =
	SettingsBlock &operator = (const SettingsBlock &sb) {
		pttLeadIn = sb.pttLeadIn;
		pttLeadOut = sb.pttLeadOut;
		convertCRLF = sb.convertCRLF;

		boxcarLength = sb.boxcarLength;
		minimumAverageDistance = sb.minimumAverageDistance;
		txMode = sb.txMode;
		txSpeed = sb.txSpeed;
		rxMode = sb.rxMode;
		rxSpeed = sb.txSpeed;

		#if CW_DATA_SOURCE == CW_SOURCE_567
		inputBounce = sb.inputBounce;
		#endif

		#if CW_DATA_SOURCE == CW_SOURCE_TSH
		I = sb.I;
		Q = sb.Q;
		maxCV = sb.maxCV;

		pllReportingGain = sb.pllReportingGain;
		pllReportingOffset = sb.pllReportingOffset;
		pllLowCorrection = sb.pllLowCorrection;

		line_input = sb.line_input;
		#endif
		return (*this);
	}
} Settings;

// flag indicating that the settings should be written
static bool settingsDirty = false;
static uint32_t lastSettingsWrite = 0;
#define SettingsInterval (10 * 1000) // msec

// the EEPROM key
static const char *EEPromKey = "CWMOD_101";

// the EEPROM access
static EEBlock eeprom(EEPromKey);

// Limits
#define MaxPttLeadIn (500)		// msec
#define MaxPttLeadOut (1000)	// msec
#define MaxWPM (60)				// WPM
#define MinWPM (5)				// WPM
#define MaxBounceInterval (1200L / MaxWPM)	// msec
#define MinBoxCarLength (8)		// elements
#define MaxBoxCarLength (65535)	// elements

#if CW_DATA_SOURCE == CW_SOURCE_TSH
#define LINE_INPUT_MAX (15)
#endif

#if CW_DATA_SOURCE == CW_SOURCE_567
#define DefaultKeyDebounceInterval (3)	// msec
#endif

// buffer limits
#ifdef TEENSYDUINO
#define InputStringLength (80)
#define CommandStringLength (80)
#else
#define InputStringLength (20)
#define CommandStringLength (20)
#endif

// control keys
#define ESC (0x1B)			// key used for all-stop
#define CtlC (0x03)			// alternative key for all-stop
#define Backspace (0x08)	// backspace key
#define Delete (0x7F)		// delete key

// CR/LF conversion handling
uint8_t lastCRLF = 0;				// holds the last character to prevent CRLF from converting twice
bool lastCharWasSpace = true;		// used to suppress extra spaces


////////////////////////////////////////////////////////////////
//
//  Support Methods
//
////////////////////////////////////////////////////////////////

//
//  parse a SIGNED integer
//
static int32_t parseInt(String argument) {
	bool negative = (argument.length() != 0 && argument[0] == '-');
	if (negative)
		argument.remove(0, 1);
	int32_t newVal = argument.toInt();
	if (negative)
		newVal = -newVal;
	return newVal;
}

//
//  ProcessElement(...) - run the decoder chain on a state change
//
static void ProcessElement(unsigned width, bool state) {
	if (!state) {
		if (width > 1000) {
			width = 1000;
		}
		if (width == 0) {
			width = 1;
		}
	}

	// build the CW element representing this pulse or pause
	CwElement cw;
	cw.Mark = (bool)(!state);	// the detector pulls the pin LOW on contact (567, button, relay, etc.)
	cw.Length = width;
	CwBuffer.Add(cw);			// TODO: check this for failure, to detect overrun
	
	// push the pulse into the timer
	if (Timing->Decode(CwBuffer, ElementBuffer)) {
		// if the timer produced elements, run those through the decoder
		byte ct = Decoder->Decode(ElementBuffer, ioBuffer, 8);
		// if that produced printable characters, print them
		if (ct > 0) {
			ioBuffer[ct] = 0;
			for (uint8_t idx = 0; idx != ct; ++idx) {
				char ch = ioBuffer[idx];
				bool wasSpace = (ch == ' ');

				if (!wasSpace || !lastCharWasSpace) {
					// print each received character
					HostSerial.print(ch);
				}

				lastCharWasSpace = wasSpace;

				// if this is BT, and echo enabled, print a CRLF (new line)
				if (echo && (ch == '=')) {
					HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					lastCharWasSpace = true;
				}
			}
		}
	}
}


#if CW_DATA_SOURCE == CW_SOURCE_TSH // Teensy audio handler
	//
	//  ProcessRxQueue() - RX audio data handler for Teensy audio library;
	//                     this is not an interrupt handler, it only processes
	//                     the samples in the audio queue...
	//
	static void ProcessRxQueue() {
		#if CW_LED_PIN != 0
		bool led = false;
		#endif
		while (inQueue.available()) {
			// read one audio frame
			int16_t *data = inQueue.readBuffer();
			int16_t * const target = data + SG_PACKET_SIZE;

			// process each sample within that frame
			while (data != target) {
				int16_t sample = *data++;

				// run the audio clock on the 44.1kHz input stream; this produces a 1kHz clock, recorded in 'audioClock'
				if (++audioClockCtr == audioClockTarget) {
					audioClockCtr = 0;
					++audioClock;
					if (++audioClockLeap == 10) {
						audioClockLeap = 0;
						audioClockTarget = 45;
					} else {
						audioClockTarget = 44;
					}
				}

				#if CW_RATE_DIVISOR != 1
				// downsample, if enabled
				if (++downsample_ctr >= CW_RATE_DIVISOR) {
					downsample_ctr = 0; // process this sample
				} else {
					continue; // skip this sample
				}
				#endif

				// do the demod and store the resulting state (inverted)
				audioState = ! detector.run(sample);

				#if EMULATE_LM567 == true
				// store the current smoothed PLL offset, to use for a tuning aid
				if (!audioState) {
					float off = pllTrackOffset + pllOffsetLPF.run(pllTrackGain * (detector.EffectivePeriod() - detector.Period()));
					if (off < 0)
						off *= pllTrackLowScale;
					pllOffset = off;
					pllOffsetActive = true;
				}
				#endif

				// if the state has changed
				if (audioState != lastState) {
					bool newState = lastState;

					// calculate the pulse width
					unsigned long width = Elapsed(audioClock, lastTime);

					// update the decoder
					ProcessElement(width, lastState);

					// if the new state is key-up, look for trailing space
					if (newState) {
						hanging = true;
					}

					// save the time and state of this edge
					lastTime = audioClock;
					lastState = audioState;
				} else {
					// handling the trailing edge of the last key-down element
					if (lastState && hanging) {
						unsigned ws = Timing->DotLength() * Timing->MinimumWordSpace();
						if (Elapsed(audioClock, lastTime) > ws) {
							ProcessElement(ws, true);
							hanging = false;
						}
					}
				}
			} // ... end process one SAMPLE

			// free the processed frame
			inQueue.freeBuffer();

			#if CW_LED_PIN != 0
			// set flag to process CW LED
			led = true;
			#endif
			#ifdef CW_TRIGRAPH
			if (!audioState) {
				if (triCtr < triTarget) {
					++triCtr;
				}
			} else {
				triCtr = 0;
			}
			#endif
		} // ... end process one FRAME

		#if CW_LED_PIN != 0
		// update the LED decode pin
		if (led) {
			digitalWrite(CW_LED_PIN, ! audioState);
		}
		#endif

		#ifdef CW_TRIGRAPH
		// paint the trigraph
		if (triCtr == triTarget) {
			int8_t tri = pllOffset;
			if (pllOffset > 127) tri = 127;
			if (pllOffset < -128) tri = -128;
			trigraph.update(tri);
			lastPllOffsetUpdate = audioClock;
		}
		#endif

		#if EMULATE_LM567 == true
		// mute the PLL offset tracking if the PLL has unlocked
		if (pllOffsetActive) {
			if (Elapsed(audioClock, lastPllOffsetUpdate) >= PLL_OFFSET_LATCH_DELAY) {
				pllOffset = 0;
				pllOffsetActive = true;

				// clear the trigraph
				#ifdef CW_TRIGRAPH
				trigraph.clear();
				#endif
			}
		}
		#endif
	}

#elif CW_DATA_SOURCE == CW_SOURCE_AA0
	//
	//  A/D Interrupt Handler -- called when new A/D value ready
	//
	ISR(ADC_vect) {
		// the A/D input var
		static int8_t sample = 0;

		//
		//  NOTE: BE VERY FAST HERE; this handler is being called
		//        thousands of times per second, and doing too
		//        much work here will cause samples to be lost.
		//

		// enable this to occasionally print the number of usec consumed per sample
		#if DEBUG_CPU_LOAD
		static uint16_t isr_count = 0;
		uint32_t isr_time = micros();
		#endif // DEBUG_CPU_LOAD

		// read the current A/D value and remove the DC offset
		sample = ADCH - ANALOG_DC_OFFSET;

		// check for clipping; store max level
		const uint8_t alevel = abs(sample);
		if (alevel > audio_max)
			audio_max = alevel;

		// do the demod and store the resulting state (inverted)
		audioState = ! detector.run(sample);

		// enable this to occasionally print the number of usec consumed per sample
		#if DEBUG_CPU_LOAD
		if (++isr_count == CW_SAMPLE_RATE) {
			isr_time = micros() - isr_time;
			HostSerial.println(isr_time);
			isr_count = 0;
		}
		#endif // DEBUG_CPU_LOAD
	}

	//
	//  Timer1 OCRB ISR - this is the timer that runs the A/D sample rate
	//      generation, which drives the receiver
	//
	ISR(TIMER1_COMPB_vect) {
		// nop -- returning from this method will trigger the A/D to start a
		//        new sample; just the same, the function must be here.
	}
#endif


//
//  ServiceSerialPortInputQueue() - empty the Serial object connected to the host port
//
static inline char ServiceSerialPortInputQueue() {
	// service the serial port (one byte at a time)
	if (HostSerial.available()) {
		// get the new byte
		char inChar = (char)HostSerial.read();
		// add it to the input string
		if (inputString.length() < InputStringLength) {
			inputString += inChar;
			inputReady = true;
		}
		return inChar;
	}
	return -1;
}


//
//  setup()
//
void setup() {
	// start the serial port
	HostSerial.begin(SerialRate);

#ifdef VERBOSE_DEBUG
	HostSerial.println();
	HostSerial.print(strcpy_P(ioBuffer, PSTR("64 = ")));
	HostSerial.println(sizeof(uint64_t));
	HostSerial.print(strcpy_P(ioBuffer, PSTR("N  = ")));
	HostSerial.println(detector.N);
	HostSerial.print(strcpy_P(ioBuffer, PSTR("w  = ")));
	HostSerial.println(detector.dist);
	HostSerial.print(strcpy_P(ioBuffer, PSTR("fr = ")));
	HostSerial.println(detector.frac.toDouble());
	HostSerial.print(strcpy_P(ioBuffer, PSTR("1m = ")));
	HostSerial.println(detector.frac_1m.toDouble());
	HostSerial.print(strcpy_P(ioBuffer, PSTR("fs = ")));
	HostSerial.println(CW_SAMPLE_RATE);
#endif
	
	// allocate strings
	cmdText.reserve(CommandStringLength);
	inputString.reserve(InputStringLength);
	inputString = strcpy_P(ioBuffer, PSTR("#VERSION;"));
	inputReady = true;

#if CMD_LED > 0
	// the 'command' LED
	pinMode(CMD_LED, OUTPUT);
	digitalWrite(CMD_LED, LOW);
	lastCmdLED = LOW;

	// send 'K' to the command LED (just a toy)
	digitalWrite(CMD_LED, HIGH);
	delay(150);
	digitalWrite(CMD_LED, LOW);
	delay(50);
	digitalWrite(CMD_LED, HIGH);
	delay(50);
	digitalWrite(CMD_LED, LOW);
	delay(50);
	digitalWrite(CMD_LED, HIGH);
	delay(150);
	digitalWrite(CMD_LED, LOW);
	delay(250);
#endif // CMD_LED > 0
	
	//
	// CW Decode LED
	//
#if CW_LED_PIN > 0
	pinMode(CW_LED_PIN, OUTPUT);
	digitalWrite(CW_LED_PIN, LOW);
#endif

	// trigraph LEDs
#ifdef CW_TRIGRAPH
	trigraph.setup();
	trigraph.test();
#endif
	
	// CW TX (key) pin
	pinMode(CW_KEY_PIN, OUTPUT);
	digitalWrite(CW_KEY_PIN, LOW);

	// CW PTT pin
#if CW_PTT_PIN > 0
	pinMode(CW_PTT_PIN, OUTPUT);
	digitalWrite(CW_PTT_PIN, LOW);
#endif
	pttState = false;

	// save initial timestamps
	lastTime = millis();
#if CMD_LED > 0
	lastRxData = lastTime;
#endif
	
	// initialize WPM and set speed to auto-track
	Timing->RxWPM(25);
	Timing->TxWPM(25);
	Timing->RxMode(SpeedAuto);
	Timing->TxMode(SpeedAuto);

	//
	// CW RX input
	//
#if CW_DATA_SOURCE == CW_SOURCE_AA0
	// turn the receiver on
	StartAudioInput(CW_SAMPLE_RATE, true);
#elif CW_DATA_SOURCE == CW_SOURCE_TSH
	// allocate audio memory
	AudioMemory(60);

	// configure the input and output filters
	uint8_t order = CW_ORDER;
	if (order > 4) order = 4;
	for (uint8_t i = 0; i != order; ++i) {
		iirIn.setBandpass(i, CW_FREQ, CW_FREQ / CW_BW);
		//iirOut.setBandpass(0, CW_FREQ, CW_FREQ / CW_BW);
	}

	uint8_t line_in = CW_INPUT_LEVEL;

#elif CW_DATA_SOURCE == CW_SOURCE_567
	// configure the RX pin for digial input from a demod IC
	pinMode(CW_RX_PIN, INPUT_PULLUP);
	KeyInput.attach(CW_RX_PIN);
	KeyInput.interval(DefaultKeyDebounceInterval);
#endif

	// read persisted settings
	SettingsBlock toRead;
	if (eeprom.read(reinterpret_cast<char*>(&toRead), sizeof(toRead)) == sizeof(toRead)) {
		// store the new values
		Settings = toRead;

		// configure the timing module
		Timing->BoxCarLength(Settings.boxcarLength);
		Timing->MinimumAverageDistance(Settings.minimumAverageDistance);
		Timing->RxWPM(Settings.rxSpeed);
		Timing->TxWPM(Settings.txSpeed);
		Timing->RxMode(Settings.rxMode);
		Timing->TxMode(Settings.txMode);

		#if CW_DATA_SOURCE == CW_SOURCE_567
		KeyInput.interval(Settings.bounceInterval);
		#endif

		#if CW_DATA_SOURCE == CW_SOURCE_TSH
		detector.SetLengthI(Settings.I);
		detector.SetLengthQ(Settings.Q);
		detector.SetMaxCV(Settings.maxCV);
		pllTrackGain = Settings.pllReportingGain;
		pllTrackOffset = Settings.pllReportingOffset;
		pllTrackLowScale = Settings.pllLowCorrection;
		line_in = Settings.line_input;
		#endif
	}

	// configure the shield for LINE INPUT, with no processing
	sgtl5000.enable();
	sgtl5000.inputSelect(AUDIO_INPUT_LINEIN);
	sgtl5000.lineInLevel(line_in);
	// configure the shield for LINE OUTPUT, with no processing
	sgtl5000.muteHeadphone();
	sgtl5000.unmuteLineout();
	sgtl5000.autoVolumeDisable();
	sgtl5000.surroundSoundDisable();
	sgtl5000.enhanceBassDisable();
	sgtl5000.eqSelect(FLAT_FREQUENCY);
	sgtl5000.adcHighPassFilterDisable();

	// start sampling
	inQueue.begin();
}


//
//  loop()
//
void loop() {
	// read the clock
	unsigned long now = millis();

	// extinguish the data LED
	#if CMD_LED > 0
	if ((lastCmdLED == HIGH) && (Elapsed(now, lastRxData) > DataLedDelay)) {
		lastCmdLED = LOW;
		digitalWrite(CMD_LED, LOW);
	}
	#endif
	
	//
	//  TX: if currently in transmit mode, service the TX queue
	//
	if (tx) {
		// if the current element has been sent...
		if (Elapsed(now, curTxStart) >= curTxLength) {
			// if there are no more elements to send, generate some more
			if (CwBuffer.Count() == 0 && TxBuffer.Count()) {
				char next;
				// encode one byte
				if (TxBuffer.Remove(next) && Decoder->Encode(ElementBuffer, &next, 1)) {
					// encode the char
					Timing->Encode(ElementBuffer, CwBuffer);
				}
			}
			
			// if there are elements to send, start sending the first one
			if (CwBuffer.Remove(cwOut)) {
				curTxStart = now;
				curTxLength = cwOut.Length;
				digitalWrite(CW_KEY_PIN, cwOut.Mark);
				if (!cwOut.Mark) {
					lastKeyDown = now;
				}
							
				#ifdef VERBOSE_DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("New Element: M=")));
				HostSerial.print(cwOut.Mark);
				HostSerial.print(strcpy_P(ioBuffer, PSTR("; L=")));
				HostSerial.println(cwOut.Length);
				#endif
			} else {
				tx = false;
#if CW_DATA_SOURCE == CW_SOURCE_AA0
				// turn the receiver back on
				StartAudioInput(CW_SAMPLE_RATE, true);
#elif CW_DATA_SOURCE == CW_SOURCE_TSH
				// start sampling
				inQueue.begin();
#endif
			}
		}
	} // if (tx)...

	// key-up (semi break-in timer expired)
	now = millis();
	if (!tx && pttState && Elapsed(now, lastKeyDown) > Settings.pttLeadOut) {
#if CW_PTT_PIN > 0
		digitalWrite(CW_PTT_PIN, LOW);
#endif
		if (echo)
			HostSerial.print(' ');
		pttState = false;
	}
	
	// if terminal/serial/USB user input ready
	if (inputReady) {
		// light the CMD LED
		#if CMD_LED > 0
		lastRxData = millis();
		if (lastCmdLED == LOW) {
			lastCmdLED = HIGH;
			digitalWrite(CMD_LED, HIGH);
		}
		#endif
		
		// for each char received from the USB...
		for (unsigned i = 0; i != inputString.length(); ++i) {
			char ch = inputString.charAt(i);
			
			//
			//  CMD: if currently in command mode...
			//
			if (cmd) {
				if (ch == '#') {
					// if a '#' sent mid-command, start over
					cmdText = "";
				} else if (ch == ';') {
					// terminate and process the command string
					cmd = false;
					bool ok = false;
					if (echo == 1) HostSerial.print(';');

					// parse the command and argument
					SplitCommand(cmdText, command, argument);
					if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOC")))) {
						//
						//  QUERY DOC - this will return a list of valid commands
						//
						if (argument.length() == 0) {
							if (echo == 1)
								HostSerial.println();
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#OK:DOC=")));
							HostSerial.print(strcpy_P(ioBuffer, PSTR("RXWPM,TXWPM,RXMODE,TXMODE,")));
							HostSerial.print(strcpy_P(ioBuffer, PSTR("PTTIN,PTTOUT,ECHO,DOT,")));
#if ( CW_DATA_SOURCE == CW_SOURCE_AA0 || CW_DATA_SOURCE == CW_SOURCE_TSH )
#if CW_DATA_SOURCE == CW_SOURCE_TSH
							HostSerial.print(strcpy_P(ioBuffer, PSTR("DEV,LINEIN,LEVEL,")));
#endif
#if EMULATE_LM567 == true
							HostSerial.print(strcpy_P(ioBuffer, PSTR("PLLI,PLLQ,PLLR,")));
							HostSerial.print(strcpy_P(ioBuffer, PSTR("PLLOFF,PLLGAIN,PLLLOW,")));
#endif
#else
							HostSerial.print(strcpy_P(ioBuffer, PSTR("BOUNCE,")));
#endif
							HostSerial.print(strcpy_P(ioBuffer, PSTR("BCL,DZ,CRLFBT,MODE,VERSION;")));
							if (echo == 1)
								HostSerial.println();

							// clear the command buffer
							cmdText = "";

							// this command is specal, skip normal handling
							continue;
						} else {
							//
							//  TODO: process individual query commands to get details
							//
							ok = false;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("RXWPM")))) {
						//
						//  QUERY/SET WPM (RX)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Timing->RxWPM());
						} else {
							int newWPM = Timing->RxWPM();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newWPM < MaxWPM)
											++newWPM;
									} break;
									case '-': {
										if (newWPM > MinWPM)
											--newWPM;
									} break;
									default: newWPM = argument.toInt(); break;
								}
							} else {
								newWPM = argument.toInt();
							}
							if (newWPM >= MinWPM && newWPM <= MaxWPM) {
								// read the current TX timing mode
								SpeedSources source = Timing->RxMode();

								// set the new speed
								Timing->RxWPM(newWPM);
								settingsDirty = true;
								cmdText = command + '=' + (int)(newWPM);

								// if the TX timing mode has changed, send a message
								if (Timing->RxMode() != source) {
									if (echo == 1)
										HostSerial.println();
									HostSerial.print(strcpy_P(ioBuffer, PSTR("#RXMODE=")));
									HostSerial.print(Timing->RxMode() == SpeedAuto ? "AUTO;" : "MANUAL;");
								}

								// success
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TXWPM")))) {
						//
						//  QUERY/SET WPM (TX)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Timing->TxWPM());
						} else {
							int newWPM = Timing->RxWPM();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newWPM < MaxWPM)
											++newWPM;
									} break;
									case '-': {
										if (newWPM > MinWPM)
											--newWPM;
									} break;
									default: newWPM = argument.toInt(); break;
								}
							} else {
								newWPM = argument.toInt();
							}
							if (newWPM >= MinWPM && newWPM <= MaxWPM) {
								// read the current TX timing mode
								SpeedSources source = Timing->TxMode();

								// set the tx speed
								Timing->TxWPM(newWPM);
								settingsDirty = true;
								cmdText = command + '=' + (int)(newWPM);

								// if the TX timing mode has changed, send a message
								if (Timing->TxMode() != source) {
									if (echo == 1)
										HostSerial.println();
									HostSerial.print(strcpy_P(ioBuffer, PSTR("#TXMODE=")));
									HostSerial.print(Timing->TxMode() == SpeedAuto ? "AUTO;" : "MANUAL;");
								}

								// success
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("CRLFBT")))) {
						//
						//  QUERY/SET CR/LF CONVERSION
						//
						//  if enabled, this will cause CR/LF to be transmitted as BT prosign
						//
						//  valid arguments are 1 for enabled, or 0 for disabled
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += Settings.convertCRLF ? "1" : "0";
						} else {
							int newcc = Settings.convertCRLF;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newcc < 1)
											++newcc; 
									} break;
									case '-': {
										if (newcc > 0)
											--newcc;
									} break;
									default: newcc = argument.toInt(); break;
								}
							} else {
								newcc = argument.toInt();
							}
							Settings.convertCRLF = newcc ? true : false;
							cmdText = command + '=' + (int)(newcc);
							settingsDirty = true;
							ok = true;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ECHO")))) {
						//
						//  QUERY/SET ECHO
						//
						//  Valid arguments:
						//    0 - echo disabled
						//    1 - echo enabled for both data and commands
						//    2 - echo enabled for data only
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(echo);
						} else {
							int newEcho = echo;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newEcho < 2)
											++newEcho; 
									} break;
									case '-': {
										if (newEcho > 0)
											--newEcho;
									} break;
									default: newEcho = argument.toInt(); break;
								}
							} else {
								newEcho = argument.toInt();
							}
							if (newEcho >= 0 && newEcho <= 2) {
								cmdText = command + '=' + (int)(newEcho);
								echo = newEcho;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTIN")))) {
						//
						//  QUERY/SET PTT LEAD-IN - argument in milliseconds
						//
						//  This is the time between when PTT is asserted and the first KEY data is transmitted
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Settings.pttLeadIn);
						} else {
							int newPTT = Settings.pttLeadIn;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newPTT < MaxPttLeadIn)
											++newPTT; 
									} break;
									case '-': {
										if (newPTT > 0)
											--newPTT;
									} break;
									default: newPTT = argument.toInt(); break;
								}
							} else {
								newPTT = argument.toInt();
							}
							if (newPTT >= 0 && newPTT <= MaxPttLeadIn) {
								Settings.pttLeadIn = newPTT;
								cmdText = command + '=' + (int)(newPTT);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTOUT")))) {
						//
						//  QUERY/SET PTT LEAD-OUT - argument in milliseconds
						//
						//  This is the time between the last KEY edge and the de-assert of PTT
						//
						//  This is the same as "QSK Delay" for semi-QSK operation
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Settings.pttLeadOut);
						} else {
							int newPTT = Settings.pttLeadOut;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newPTT < MaxPttLeadOut)
											++newPTT; 
									} break;
									case '-': {
										if (newPTT > 0)
											--newPTT;
									} break;
									default: newPTT = argument.toInt(); break;
								}
							} else {
								newPTT = argument.toInt();
							}
							if (newPTT >= 0 && newPTT <= MaxPttLeadOut) {
								Settings.pttLeadOut = newPTT;
								cmdText = command + '=' + (int)(newPTT);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOT")))) {
						//
						//  QUERY DOT LENGTH - argument in milliseconds
						//
						ok = true;
						cmdText = command;
						cmdText += '=';
						cmdText += (int)(Timing->DotLength());
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("GAP")))) {
						//
						//  QUERY SAFETY GAP
						//
						ok = true;
						cmdText = command;
						cmdText += '=';
						cmdText += (int)(Timing->SafetyGap());
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("RXMODE")))) {
						//
						//  ENABLE/DISABLE RX WPM AUTO-TRACK
						//
						//  Valid arguments are AUTO and MANUAL
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Timing->RxMode() == SpeedAuto ? "AUTO" : "MANUAL";
						} else {
							if (argument.equalsIgnoreCase("AUTO")) {
								Timing->RxMode(SpeedAuto);
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase("MANUAL")) {
								Timing->RxMode(SpeedManual);
								settingsDirty = true;
								ok = true;
							} else if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+':
									case '-':
										Timing->RxMode((Timing->RxMode() == SpeedAuto) ? SpeedManual : SpeedAuto);
										cmdText = command + '=';
										cmdText += Timing->RxMode() == SpeedAuto ? "AUTO" : "MANUAL";
										settingsDirty = true;
										ok = true;
										break;
								}
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TXMODE")))) {
						//
						//  ENABLE/DISABLE TX WPM AUTO-TRACK
						//
						//  This will cause the TXWPM to follow the RXWPM when RXMODE = AUTO
						//
						//  Valid arguments are AUTO and MANUAL
						//
						ok = true;
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Timing->TxMode() == SpeedAuto ? "AUTO" : "MANUAL";
						} else {
							if (argument.equalsIgnoreCase("AUTO")) {
								Timing->TxMode(SpeedAuto);
								settingsDirty = true;
								ok = true;
							} else if (argument.equalsIgnoreCase("MANUAL")) {
								Timing->TxMode(SpeedManual);
								settingsDirty = true;
								ok = true;
							} else if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+':
									case '-':
										Timing->TxMode((Timing->TxMode() == SpeedAuto) ? SpeedManual : SpeedAuto);
										cmdText = command + '=';
										cmdText += Timing->TxMode() == SpeedAuto ? "AUTO" : "MANUAL";
										settingsDirty = true;
										ok = true;
										break;
								}
							}
						}
#if CW_DATA_SOURCE == CW_SOURCE_AA0 || CW_DATA_SOURCE == CW_SOURCE_TSH
#if EMULATE_LM567 == true
#if CW_DATA_SOURCE == CW_SOURCE_TSH
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DEV")))) {
						//
						//  GET VALUE OF PLL CONTROL VOLTAGE
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += pllOffset;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLOFF")))) {
						//
						//  GET/SET PLL TRACK OFFSET
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int32_t)(pllTrackOffset);
						} else {
							int newValue = pllTrackOffset;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newValue < MaxPttLeadOut)
											++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							pllTrackOffset = newValue;
							cmdText = command + '=' + (int)(pllTrackOffset);
							settingsDirty = true;
							ok = true;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLGAIN")))) {
						//
						//  GET/SET PLL TRACK GAIN/SCALE
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int32_t)(pllTrackGain);
						} else {
							int newValue = pllTrackGain;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newValue < MaxPttLeadOut)
											++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue > 0) {
								pllTrackGain = newValue;
								cmdText = command + '=' + (int)(pllTrackGain);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLLOW")))) {
						//
						//  GET/SET PLL TRACKING LOW SCALE CORRECTION
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (smoother_t)(pllTrackLowScale);
						} else {
							smoother_t newVal = argument.toFloat();
							if (newVal > 0) {
								pllTrackLowScale = newVal;
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LINEIN")))) {
						//
						//  QUERY/SET LINE IN VOLUME
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += Settings.line_input;
						} else {
							unsigned long newLevel = argument.toInt();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										newLevel = Settings.line_input;
										if (Settings.line_input < LINE_INPUT_MAX) {
											newLevel = Settings.line_input + 1;
										}
									} break;
									case '-': {
										newLevel = Settings.line_input;
										if (Settings.line_input > 0) {
											newLevel = Settings.line_input - 1;
										}
									} break;
								}
							}
							if (newLevel >= 0 && newLevel <= LINE_INPUT_MAX) {
								Settings.line_input = newLevel;
								sgtl5000.lineInLevel(newLevel);
								cmdText = command + '=' + Settings.line_input;
								settingsDirty = true;
								ok = true;
							}
						}
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLI")))) {
						//
						//  GET/SET PLL I INTEGRATOR LENGTH
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(detector.GetLengthI());
						} else {
							int newValue = detector.GetLengthI();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue > 0) {
								detector.SetLengthI(newValue);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLQ")))) {
						//
						//  GET/SET PLL Q INTEGRATOR LENGTH
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(detector.GetLengthQ());
						} else {
							int newValue = detector.GetLengthQ();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue > 0) {
								detector.SetLengthQ(newValue);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PLLR")))) {
						//
						//  GET/SET PLL VCO RAIL
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(detector.GetMaxCV());
						} else {
							int newValue = detector.GetMaxCV();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue > 0) {
								detector.SetMaxCV(newValue);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("I")))) {
						//
						//  GET VALUE OF 'I' INTEGRATOR
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += detector.I();
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("Q")))) {
						//
						//  GET VALUE OF 'Q' INTEGRATOR
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += detector.Q();
						}
#else
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DL")))) {
						// return the current detector internal level
						ok = true;
						cmdText = command;
						cmdText += '=';
						cmdText += detector.level();
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DS")))) {
						// return the current detector output
						ok = true;
						cmdText = command;
						cmdText += '=';
						cmdText += audioState ? strcpy_P(ioBuffer, PSTR("HIGH")) : strcpy_P(ioBuffer, PSTR("LOW"));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVEL")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText = command;
						cmdText += '=';

						//
						// present the peak level as a percentage
						//
						#if CW_DATA_SOURCE == CW_SOURCE_AA0
						cmdText += 100 * static_cast<uint16_t>(audio_max) / 128;
						// reset the peak
						audio_max = 0;
						#elif CW_DATA_SOURCE == CW_SOURCE_TSH
						// read the peak and normalize to %; the peakIn will reset automatically
						cmdText += static_cast<uint16_t>(100.0 * peakIn.read());
						#endif
#else
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BOUNCE")))) {
						//
						//  GET/SET CW INPUT DEBOUNCE INTERVAL - argument in milliseconds
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(KeyInput.interval());
						} else {
							int newValue = KeyInput.interval();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newValue < MaxBounceInterval)
											++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument); break;
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue >= 0 && newValue <= MaxBounceInterval) {
								KeyInput.interval(newValue);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("BCL")))) {
						//
						//  GET/SET CW SPEED TRACKING BOXCAR LENGTH
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Timing->BoxCarLength());
						} else {
							int newValue = Timing->BoxCarLength();
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newValue < MaxBoxCarLength)
											++newValue; 
									} break;
									case '-': {
										if (newValue > MinBoxCarLength)
											--newValue;
									} break;
									default: newValue = parseInt(argument);
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue >= MinBoxCarLength && newValue <= MaxBoxCarLength) {
								Timing->BoxCarLength(newValue);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DZ")))) {
						//
						//  GET/SET CW SPEED TRACKING DEAD ZONE (percent)
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText = command;
							cmdText += '=';
							cmdText += (int)(Timing->MinimumAverageDistance() * 100);
						} else {
							unsigned newValue = Timing->MinimumAverageDistance() * 100;
							if (argument.length() == 1) {
								switch (argument.charAt(0)) {
									case '+': {
										if (newValue < 100)
											++newValue; 
									} break;
									case '-': {
										if (newValue > 0)
											--newValue;
									} break;
									default: newValue = parseInt(argument);
								}
							} else {
								newValue = parseInt(argument);
							}
							if (newValue >= 0 && newValue <= 100) {
								Timing->MinimumAverageDistance((float)newValue / 100.0);
								cmdText = command + '=' + (int)(newValue);
								settingsDirty = true;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("EC")))) {
						//
						//  CLEAR EEPROM
						//
						if (argument.length() == 1 && argument.charAt(0) == '1') {
							eeprom.clear();
							settingsDirty = false;
							ok = true;
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("CWTSTATE")))) {
						//
						//  PRINT CURRENT CW TIMING STATE - this is for debugging only; do not add to DOC string
						//
						CwTimingLogic *morse = (CwTimingLogic*)(Timing);
						if (morse) {
							ok = true;
							CwTimingState state = morse->GetState();
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,dotlength=")));
							HostSerial.print(state.DotLength);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,txdotlen=")));
							HostSerial.print(state.TxDotLength);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,rxdotlen=")));
							HostSerial.print(state.RxDotLength);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,bcavg=")));
							HostSerial.print(state.BoxCarAverage);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,minavg=")));
							HostSerial.print(state.MinimumAverageDistance);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,safety=")));
							HostSerial.print(state.SafetyGap);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,txsrc=")));
							HostSerial.print(state.TxSpeedSource);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,rxsrc=")));
							HostSerial.print(state.RxSpeedSource);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,maxdotlen=")));
							HostSerial.print(state.MaximumDotLength);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,minwordsp=")));
							HostSerial.print(state.MinimumWordSpace);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,txdashmult=")));
							HostSerial.print(state.TxDashMultiplier);
							HostSerial.println(';');
							HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug,txwordmult=")));
							HostSerial.print(state.TxWordMultiplier);
							HostSerial.println(';');
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("MODE")))) {
						//
						//  QUERY OP MODE
						//
						ok = true;
						cmdText = command;
						cmdText += "=CW";
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("VERSION")))) {
						//
						//  QUERY VERSION NUMBER
						//
						ok = true;
						cmdText = command;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR(VERSION_STRING));
					} else if (command.length() == 0) {
						//
						//  empty command
						//
						ok = true;
					}

					//
					//  TODO: process future command contents
					//

					// respond with #OK or #ERR
					if (echo == 1) {
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					}
					if (command.length() == 0) {
						HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
					} else {
						HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK:")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
						HostSerial.print(cmdText);
					}
					HostSerial.print(';');
					if (echo == 1) {
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
						lastCharWasSpace = true;
					}
					cmdText = "";
				} else {
					if (isAlpha(ch) || isDigit(ch) || isPunct(ch) || (ch == ' ')) {
						// add the char to the command string
						if (cmdText.length() < CommandStringLength)
							cmdText += ch;
						
						// echo the char, if enabled
						if (echo == 1) {
							HostSerial.print(ch);
						}
					}
				}
				continue;
			} // if (cmd)....
			
			// start command mode??
			if ((!cmd) && (ch == '#')) {
				cmd = true;
				if (echo == 1) HostSerial.print('#');
				continue;
			}

			// user hard-stop?
			if (ch == ESC || ch == CtlC) {
				digitalWrite(CW_KEY_PIN, LOW);
#if CW_PTT_PIN > 0
				digitalWrite(CW_PTT_PIN, LOW);
#endif
				TxBuffer.Clear();
				CwBuffer.Clear();
				ElementBuffer.Clear();
				pttState = false;
				tx = false;
#if CW_DATA_SOURCE == CW_SOURCE_AA0
				// turn the receiver back on
				StartAudioInput(CW_SAMPLE_RATE, true);
#elif CW_DATA_SOURCE == CW_SOURCE_TSH
				// turn the receiver back on
				inQueue.begin();
#endif
				continue;
			}

			// handle backspace and delete as backspace
			if (ch == Backspace || ch == Delete) {
				if (TxBuffer.Count() > 0) {
					char dummy;
					TxBuffer.RemoveLast(dummy);
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(BACKSPACE_STRING)));
				}
				continue;
			}

			// echo the char, if enabled
			if (echo) {
				// if this is BT, and echo enabled, print a CRLF (new line)
				switch (ch) {
					case '=': {
						// echo the '=' plus a CRLF
						HostSerial.print(ch);
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
						lastCharWasSpace = true;
					} break;
					case '\r':
					case '\n': {
						// print the CR or LF
						HostSerial.print(ch);
						lastCharWasSpace = true;
					} break;
					default: {
						// skip all control characters other than 
						if (iscntrl(ch)) {
							continue;
						}

						// skip all high-order ASCII characters
						if (ch >= 128) {
							continue;
						}

						// echo all others
						HostSerial.print(ch);
					} break;
				}
			}

			// upper-case the char to search for a TX pattern
			ch = toupper(ch);

			// perform CRLF conversion
			switch (ch) {
				case '\r':
				case '\n':
					// if CRLF conversion enabled, convert up to one CR or LF into BT prosign
					if (Settings.convertCRLF && ((lastCRLF == 0) || (lastCRLF == ch))) {
						lastCRLF = ch;
						ch = '=';
					}
					break;
				default:
					lastCRLF = 0;
					break;
			}

			// switch to TX mode, if needed
			if (!tx) {
				CwBuffer.Clear();
				ElementBuffer.Clear();
			}
			
			#ifdef VERBOSE_DEBUG
			HostSerial.print(strcpy(ioBuffer, PSTR("Send: ")));
			HostSerial.println(ch);
			#endif

			// if we are already in TX mode, just queue up the character
			if (tx) {
				TxBuffer.Add(ch);
			} else if ((!isspace(ch)) && Decoder->Encode(ElementBuffer, &ch, 1)) {
				// otherwise, generate characters, and go into TX mode
				if (Timing->Encode(ElementBuffer, CwBuffer)) {
					if (CwBuffer.Remove(cwOut)) {
#if CW_DATA_SOURCE == CW_SOURCE_AA0
						// turn the receiver off
						StopAudioInput();
#elif CW_DATA_SOURCE == CW_SOURCE_TSH
						// turn the receiver off
						inQueue.end();
						inQueue.clear();
						pllOffset = 0;
#ifdef CW_TRIGRAPH
						// clear the tuning indicator LEDs
						trigraph.clear();
#endif
#if CW_LED_PIN != 0
						// clear the CW RX data LED
						digitalWrite(CW_LED_PIN, LOW);
#endif
#endif
#if CW_PTT_PIN > 0
						// PTT lead-in
						digitalWrite(CW_PTT_PIN, HIGH);
						#if USE_PTT_DELAY
						delay(Settings.pttLeadIn);
						#else
						unsigned long pttStart = millis();
						while (Elapsed(millis(), pttStart) <= Settings.pttLeadIn) {
							// keep the RX queue from overrunning
							char ch = ServiceSerialPortInputQueue();

							// if hard-stop was pressed, just exit now so those handlers can run
							if (ch == ESC || ch == CtlC) {
								break;
							}
						}
						#endif
#endif
						pttState = true;

						// start sending
						tx = true;
						curTxStart = millis();
						curTxLength = cwOut.Length;
						digitalWrite(CW_KEY_PIN, cwOut.Mark);
						if (!cwOut.Mark) {
							lastKeyDown = curTxStart;
						}
						
						#ifdef VERBOSE_DEBUG
						HostSerial.print(strcpy_P(ioBuffer, PSTR("New Element: M=")));
						HostSerial.print(cwOut.Mark);
						HostSerial.print(strcpy_P(ioBuffer, PSTR("; L=")));
						HostSerial.println(cwOut.Length);
						#endif
					}
				}
			} // !tx
		}

		// clear the string:
		inputString = "";
		inputReady = false;
	}

	//
	//  RX: read data from input pin
	//
	if (!tx) {
#if CW_DATA_SOURCE == CW_SOURCE_TSH
		// run the algorithm on sound card input
		ProcessRxQueue();
#else
		// if the state has changed
#if CW_DATA_SOURCE == CW_SOURCE_AA0
		if (audioState != lastState) {
			bool newState = lastState;
#else
		if (KeyInput.update()) {
			bool newState = KeyInput.read();
#endif
			now = millis();
			
			// calculate the pulse width
			unsigned long width = Elapsed(now, lastTime);

#if CW_LED_PIN != 0
#if CW_DATA_SOURCE == CW_SOURCE_AA0
			digitalWrite(CW_LED_PIN, newState);
#else
			digitalWrite(CW_LED_PIN, !newState);
#endif
#endif
			// update the decoder
			ProcessElement(width, lastState);
			
			// if the new state is key-up, look for trailing space
			if (newState) {
				hanging = true;
			}

			lastTime = now;		
#if CW_DATA_SOURCE == CW_SOURCE_AA0
			lastState = audioState;
#else
			lastState = newState;
#endif
		} else {
			// handling the trailing edge of the last key-down element
			if (lastState && hanging) {
				now = millis();
				unsigned ws = Timing->DotLength() * Timing->MinimumWordSpace();
				if (Elapsed(now, lastTime) > ws) {
					ProcessElement(ws, true);
					hanging = false;
				}
			}
		}
#endif // CW_DATA_SOURCE
	} // if !tx

	// update the EEPROM if settings have changed
	uint32_t ms = millis();
	if (settingsDirty) {
		if (Elapsed(ms, lastSettingsWrite) > SettingsInterval) {
			// copy over new settings
			Settings.boxcarLength = Timing->BoxCarLength();
			Settings.minimumAverageDistance = Timing->MinimumAverageDistance();
			Settings.rxSpeed = Timing->RxWPM();
			Settings.txSpeed = Timing->TxWPM();
			Settings.rxMode = Timing->RxMode();
			Settings.txMode = Timing->TxMode();

			#if CW_DATA_SOURCE == CW_SOURCE_567
			Settings.bounceInterval = KeyInput.interval();
			#endif

			#if CW_DATA_SOURCE == CW_SOURCE_TSH
			Settings.I = detector.GetLengthI();
			Settings.Q = detector.GetLengthQ();
			Settings.maxCV = detector.GetMaxCV();
			Settings.pllReportingGain = pllTrackGain;
			Settings.pllReportingOffset = pllTrackOffset;
			Settings.pllLowCorrection = pllTrackLowScale;
			#endif

			// update the EEPROM
			eeprom.write(reinterpret_cast<char*>(&Settings), sizeof(Settings));
			settingsDirty = false;
			lastSettingsWrite = ms;
		}
	}

	// service the host port
	ServiceSerialPortInputQueue();
} // loop()

// EOF
