/*
 *
 *
 *    CwTimingLogic.h
 *
 *    Translates detected tone pulses into morse elements.
 *
 *    License: GNU General Public License Version 3.0.
 *
 *    Copyright (C) 2014-2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

#ifndef __CW_TIMING_H
#define __CW_TIMING_H

#include "CircularBuffer.h"
#include "Interfaces.h"
#include "Elements.h"
#include <float.h>

//
//  This is the default threshold between dot (1.0) and dash (3.0)
//
//     2.0 is the arithmetic mean between dot and dash at 3:1 weighting
//     1.7321 is the geometric mean between dot and dash at 3:1 weighting
//
#define DefaultMaximumDotLength (1.7321)

//
//  This is the default threshold between dot space and dash space
//
//     2.0 is the arithmetic mean between dot and dash at 3:1 weighting
//     1.7321 is the geometric mean between dot and dash at 3:1 weighting
//
//  Since dot-space is almost always longer than dot (because of wave shaping
//  and PLL lock vs. unlock times, this should be >= DefaultMaximumDotLength.
//
#define DefaultMaximumDotSpaceLength (2.5625)

//
//  This is the default boxcar length used when none is specified.
//
#define DefaultBoxCarLength (32)

//
//  This is the default (normalized) dot length used when none is specified.
//
#define DefaultDotLength (1.0)

//
//  This is the default size of a proper dash with respect to a dot
//
#define DefaultTxDashMultiplier (3.0)

//
//  This is the default safety gap (boxcar updates must be average +/- this times current mean)
//
#define DefaultMinimumAverageDistance (0.35)

//
//  This is the minimum space betweeen words, measured in dot spaces.
//
#define DefaultMinimumWordSpace (6)

//
//  WPM limits for automatic speed tracking
//
#define AutoSpeedMinWPM (10)
#define AutoSpeedMaxWPM (45)

//
//  Translate min/max WPM to dot lengths, used by the algorithm
//
#define AutoSpeedMinimumDotLength (1200 / AutoSpeedMaxWPM)
#define AutoSpeedMaximumDotLength (1200 / AutoSpeedMinWPM)

namespace KK5JY {
	namespace CW {
		struct CwTimingState {
			/// <summary>
			/// The current average dot length.
			/// </summary>
			float DotLength;

			/// <summary>
			/// The current average dot length used for TX.
			/// </summary>
			float TxDotLength;

			/// <summary>
			/// The current average dot length used for RX.
			/// </summary>
			float RxDotLength;

			/// <summary>
			/// The current boxcar average.
			/// </summary>
			float BoxCarAverage;

			/// <summary>
			/// The minimum average distance as a percentage.
			/// </summary>
			float MinimumAverageDistance;

			/// <summary>
			/// The current gap, computed as per MinimumAverageDistance.  This
			/// is the value actually used on a per-element basis.  The value
			/// of MinimumAverageDistance is the user interface only.
			/// </summary>
			unsigned SafetyGap;

			/// <summary>
			/// Sets the TX speed source.
			/// </summary>
			SpeedSources TxSpeedSource;

			/// <summary>
			/// Sets the RX speed source.
			/// </summary>
			SpeedSources RxSpeedSource;

			/// <summary>
			/// The maximum length for a dot or dot-space, as a multiple of the current average dot length.
			/// </summary>
			float MaximumDotLength;

			/// <summary>
			/// The minimum length for a word space, as a multiple of the current average dot length.
			/// </summary>
			float MinimumWordSpace;

			/// <summary>
			/// How many times the dot space is the dash space (on transmit)
			/// </summary>
			uint8_t TxDashMultiplier;

			/// <summary>
			/// How many times the dot space is the word space (on transmit)
			/// </summary>
			uint8_t TxWordMultiplier;
		};

		/// <summary>
		/// Translates detected elements into a logical symbol stream.
		/// </summary>
		class CwTimingLogic : public TimingLogic {
			private:
				/// <summary>
				/// The current average dot length.
				/// </summary>
				float m_DotLength;

				/// <summary>
				/// The correction applied to the dot length when auto-tracking speed.
				/// </summary>
				float m_DotCorrection;

				/// <summary>
				/// The current average dot length used for TX.
				/// </summary>
				float m_TxDotLength;

				/// <summary>
				/// The current average dot length used for RX.
				/// </summary>
				float m_RxDotLength;

				/// <summary>
				/// The boxcar array of previous mark lengths; used to compute the dot (element) size.
				/// </summary>
				float *m_BoxCar;

				/// <summary>
				/// The current boxcar sum.
				/// </summary>
				float m_BoxCarSum;

				/// <summary>
				/// The current boxcar average.
				/// </summary>
				float m_BoxCarAverage;

				/// <summary>
				/// The minimum average distance as a percentage.
				/// </summary>
				float m_MinimumAverageDistance;

				/// <summary>
				/// The index of the next item to write into the boxcar.
				/// </summary>
				uint8_t m_BoxCarIndex;

				/// <summary>
				/// The size of the boxcar array.
				/// </summary>
				uint8_t m_BoxCarSize;

				/// <summary>
				/// The current gap, computed as per MinimumAverageDistance.  This
				/// is the value actually used on a per-element basis.  The value
				/// of MinimumAverageDistance is the user interface only.
				/// </summary>
				unsigned m_SafetyGap;

				/// <summary>
				/// Sets the TX speed source.
				/// </summary>
				SpeedSources m_TxSpeedSource;

				/// <summary>
				/// Sets the RX speed source.
				/// </summary>
				SpeedSources m_RxSpeedSource;

				/// <summary>
				/// The maximum length for a dot, as a multiple of the current average dot length.
				/// </summary>
				float m_MaximumDotLength;

				/// <summary>
				/// The maximum length for a dot-space, as a multiple of the current average dot length.
				/// </summary>
				float m_MaximumDotSpaceLength;

				/// <summary>
				/// The minimum length for a word space, as a multiple of the current average dot length.
				/// </summary>
				float m_MinimumWordSpace;

				/// <summary>
				/// How many times the dot space is the dash space (on transmit)
				/// </summary>
				uint8_t m_TxDashMultiplier;

				/// <summary>
				/// How many times the dot space is the word space (on transmit)
				/// </summary>
				uint8_t m_TxWordMultiplier;

			public:
				/// <summary>
				/// Construct a new timing object.
				/// </summary>
				CwTimingLogic(uint8_t bcLength = DefaultBoxCarLength, float dotLength = DefaultDotLength);

				//
				//  virtual dtor
				//
				virtual ~CwTimingLogic() {
					/* nop */
				}

			public: // properties
				/// <summary>
				/// Return the current boxcar length.
				/// </summary>
				uint8_t BoxCarLength() const { return m_BoxCarSize; }

				/// <summary>
				/// Set a new boxcar length.
				/// </summary>
				void BoxCarLength (unsigned bcLength);

				/// <summary>
				/// Return the current average dot length from the tracker.
				/// </summary>
				float DotLength() const { return m_DotLength; }

				/// <summary>
				/// Return the correction applied to dot length calculation when auto-trackng speed.
				/// This can corect for the difference between rising/falling PLL switching times,
				/// pulse shaping, etc.
				/// </summary>
				float DotCorrection() const { return m_DotCorrection; }

				/// <summary>
				/// Return the current safety gap.
				/// </summary>
				float SafetyGap() const { return m_SafetyGap; }

				/// <summary>
				/// Estimate the current RX WPM based on the average dot length.
				/// </summary>
				float RxWPM() const { return 1200 / m_RxDotLength; }

				/// <summary>
				/// Estimate the current RX WPM based on the average dot length.
				/// </summary>
				float TxWPM() const { return 1200 / m_TxDotLength; }

				/// <summary>
				/// The minimum length for a word space, as a multiple of the current average dot length.
				/// </summary>
				float MinimumWordSpace() const { return m_MinimumWordSpace; }

				/// <summary>
				/// Set the new minimum word space.
				/// </summary>
				void MinimumWordSpace(float value) { m_MinimumWordSpace = value; }

				/// <summary>
				/// The distance, as a fraction of the average, that a new sample's pulse width must
				/// be in order to be counted in the average.  This prevents the average from collapsing
				/// in on itself when a long string of dots or dashes is encountered.
				/// </summary>
				float MinimumAverageDistance() const { return m_MinimumAverageDistance; }

				/// <summary>
				/// Set the new minimum average distance (see above).
				/// </summary>
				void MinimumAverageDistance(float mad);

				/// <summary>
				/// Set the RX WPM, and drop into manual RX mode.
				/// </summary>
				void RxWPM(int wpm);

				/// <summary>
				/// Set the TX WPM, and drop into manual TX mode.
				/// </summary>
				void TxWPM(int wpm);

				/// <summary>
				/// Set the correction applied to dot length calculation when auto-trackng speed.
				/// This can corect for the difference between rising/falling PLL switching times,
				/// pulse shaping, etc.
				/// </summary>
				void DotCorrection(float corr) { m_DotCorrection = corr; }

				//
				//  Get the RX speed source.
				//
				SpeedSources RxMode() const { return m_RxSpeedSource; }

				//
				//  Get the TX speed source.
				//
				SpeedSources TxMode() const { return m_TxSpeedSource; }

				//
				//  Set the RX speed source.
				//
				void RxMode(SpeedSources src) { m_RxSpeedSource = src; }

				//
				//  Set the TX speed source.
				//
				void TxMode(SpeedSources src) { m_TxSpeedSource = src; }

				//
				//  Get the TX dash multiplier.
				//
				uint8_t DashMultiplier() const { return m_TxDashMultiplier; }

				//
				//  Set the TX dash multiplier.
				//
				void DashMultiplier(uint8_t value) { m_TxDashMultiplier = value; }

				//
				//  Get the TX word multiplier.
				//
				uint8_t WordMultiplier() const { return m_TxWordMultiplier; }

				//
				//  Set the TX word multiplier.
				//
				void WordMultiplier(uint8_t value) { m_TxWordMultiplier = value; }

				//
				//  Get the maximum dot length
				//
				float MaximumDotLength() const { return m_MaximumDotLength; }

				//
				//  Get the maximum dot-space length
				//
				float MaximumDotSpaceLength() const { return m_MaximumDotSpaceLength; }

				//
				//  Set the minimum dot length
				//
				void MaximumDotLength(float value) { m_MaximumDotLength = value; }

				//
				//  Get the entire timing state as a block; mostly for debugging
				//
				CwTimingState GetState() const;

			public: // methods
				/// <summary>
				/// Do the decoding.
				/// </summary>
				/// <param name="raw">The raw element data.</param>
				/// <param name="result">A decoded element symbol stream.</param>
				/// <returns>True if a word space was added to the result, indicating data ready to decode.</returns>
				bool Decode(CircularBuffer<CwElement> &raw, CircularBuffer<DataElement> &result);

				/// <summary>
				/// Do the decoding.
				/// </summary>
				int Encode(CircularBuffer<DataElement> &txBuffer, CircularBuffer<CwElement> &cwBuffer);

			private:
				/// <summary>
				/// Allocate the boxcar space.
				/// </summary>
				void AllocateBoxCar(int bcLength);

				/// <summary>
				/// Initialize the boxcar with a specific dot length
				/// </summary>
				void InitializeBoxCar(int dotLength);
		};

		inline CwTimingLogic::CwTimingLogic(uint8_t bcLength, float dotLength) : m_BoxCar(0) {
			// set some reasonable default timing limits
			m_MaximumDotLength = DefaultMaximumDotLength;
			m_MaximumDotSpaceLength = DefaultMaximumDotSpaceLength;
			m_TxDashMultiplier = DefaultTxDashMultiplier;
			m_MinimumWordSpace = DefaultMinimumWordSpace;
			m_TxWordMultiplier = 5; // 7 - one dot space on each end

			// set the default minimum average distance
			m_MinimumAverageDistance = DefaultMinimumAverageDistance;

			// initialize the boxcar
			AllocateBoxCar(bcLength);
			InitializeBoxCar(dotLength);

			// set speed sources
			m_TxSpeedSource = SpeedAuto;
			m_RxSpeedSource = SpeedAuto;

			// set the default correction
			m_DotCorrection = 3.0;
		}

		inline bool CwTimingLogic::Decode(CircularBuffer<CwElement> &raw, CircularBuffer<DataElement> &result) {
			bool space = false;
			while (raw.Count() != 0) {
				// if the output is full, drop an old sample to make room for a new one
				if (result.Full())
					result.RemoveItems(1);

				// pop the first element from the input buffer
				CwElement element;
				raw.Remove(element);

				// update the element length average
				if (element.Mark) {
					if (element.Length > (m_BoxCarAverage * 4)) {
						element.Length = m_BoxCarAverage * 4;
					}
					if (element.Length < (m_BoxCarAverage / 4)) {
						element.Length = m_BoxCarAverage / 4;
					}
					if (element.Length < (m_BoxCarAverage - m_SafetyGap) || element.Length > (m_BoxCarAverage + m_SafetyGap)) {
						m_BoxCarSum -= m_BoxCar[m_BoxCarIndex];
						m_BoxCarSum += element.Length;
						m_BoxCar[m_BoxCarIndex] = element.Length;
						m_BoxCarIndex = (m_BoxCarIndex + 1) % m_BoxCarSize;

						//
						//  Compute average dot length...
						//
						//  Since this is an average of all of the elements, the
						//  overall average should be close to the midpoint
						//  between dot and dash lengths.  One half of that should
						//  be roughly the dot length.
						//
						m_BoxCarAverage = m_BoxCarSum / m_BoxCarSize;
						m_DotLength = (m_BoxCarAverage / 2.0) + m_DotCorrection;
						// clamp dot length to reasonable values
						if (m_DotLength < AutoSpeedMinimumDotLength) {
							m_DotLength = AutoSpeedMinimumDotLength;
						} else if (m_DotLength > AutoSpeedMaximumDotLength) {
							m_DotLength = AutoSpeedMaximumDotLength;
						}
						m_SafetyGap = m_MinimumAverageDistance * m_BoxCarAverage;
					}
				}

				// copy detected speed to RX speed if auto-enabled
				if (m_RxSpeedSource == SpeedAuto) {
					m_RxDotLength = m_DotLength;
				}
				// copy detected speed to TX speed if auto-enabled
				if (m_TxSpeedSource == SpeedAuto) {
					m_TxDotLength = m_DotLength;
				}

				// now decode the specific element type
				if (element.Mark) {
					if (element.Length <= (m_MaximumDotLength * m_RxDotLength)) {
						// dot
						result.Add(Dot);
					} else {
						// long elements
						result.Add(Dash);
					}
				} else { // space
					if (element.Length <= (m_MaximumDotSpaceLength * m_RxDotLength)) {
						// dot space
						result.Add(DotSpace);
					} else if (element.Length >= (m_MinimumWordSpace * m_RxDotLength)) {
						// word space
						result.Add(WordSpace);
						space = true;
					} else {
						// long elements
						result.Add(element.Mark ? Dash : DashSpace);
						space = true;
					}
				}
			}

			return space;
		}

		inline int CwTimingLogic::Encode(CircularBuffer<DataElement> &txBuffer, CircularBuffer<CwElement> &cwBuffer) {
			DataElement el;
			CwElement cw;
			const int dashLen = m_TxDotLength * m_TxDashMultiplier;
			const int wordLen = m_TxDotLength * m_TxWordMultiplier;
			int result = 0;

			while (txBuffer.Remove(el)) {
				switch (el) {
					case Dot:
						cw.Mark = true;
						cw.Length = m_TxDotLength;
						cwBuffer.Add(cw);
						++result;
						break;
					case Dash:
						cw.Mark = true;
						cw.Length = dashLen;
						cwBuffer.Add(cw);
						++result;
						break;
					case DotSpace:
						cw.Mark = false;
						cw.Length = m_TxDotLength;
						cwBuffer.Add(cw);
						++result;
						break;
					case DashSpace:
						cw.Mark = false;
						cw.Length = dashLen;
						cwBuffer.Add(cw);
						++result;
						break;
					case WordSpace:
						cw.Mark = false;
						cw.Length = wordLen;
						cwBuffer.Add(cw);
						++result;
						break;
					default:
						// nop
						break;
				}
			}
			return result;
		}

		inline void CwTimingLogic::AllocateBoxCar(int bcLength) {
			m_BoxCarSize = bcLength;
			if (m_BoxCar != 0) {
				delete [] m_BoxCar;
			}
			m_BoxCar = new float[m_BoxCarSize];
		}

		inline void CwTimingLogic::InitializeBoxCar(int dotLength) {
			m_BoxCarSum = 0;
			for (int i = 0; i != m_BoxCarSize; ++i) {
				m_BoxCar[i] = dotLength;
				m_BoxCarSum += dotLength;
			}
			m_BoxCarAverage = m_BoxCarSum / m_BoxCarSize;
			m_DotLength = (m_BoxCarAverage / 2) + m_DotCorrection;
			m_SafetyGap = m_MinimumAverageDistance * m_BoxCarAverage;
			m_BoxCarIndex = 0;
		}

		inline void CwTimingLogic::BoxCarLength (unsigned bcLength) {
			int dl = m_DotLength;
			AllocateBoxCar(bcLength);
			InitializeBoxCar(dl);
		}

		inline void CwTimingLogic::MinimumAverageDistance(float mad) {
			m_MinimumAverageDistance = mad;
			InitializeBoxCar(m_DotLength);
		}

		inline void CwTimingLogic::RxWPM(int wpm) {
			int newLength = 1200 / wpm;
			m_RxSpeedSource = SpeedManual;
			m_RxDotLength = newLength;
			InitializeBoxCar(newLength * 2); // average should be double the dot length
		}

		inline void CwTimingLogic::TxWPM(int wpm) {
			int newLength = 1200 / wpm;
			m_TxSpeedSource = SpeedManual;
			m_TxDotLength = newLength;
		}

		inline CwTimingState CwTimingLogic::GetState() const {
			CwTimingState result;
			result.DotLength = this->m_DotLength;
			result.TxDotLength = this->m_TxDotLength;
			result.RxDotLength = this->m_RxDotLength;
			result.BoxCarAverage = this->m_BoxCarAverage;
			result.MinimumAverageDistance = this->m_MinimumAverageDistance;
			result.SafetyGap = this->m_SafetyGap;
			result.TxSpeedSource = this->m_TxSpeedSource;
			result.RxSpeedSource = this->m_RxSpeedSource;
			result.MaximumDotLength = this->m_MaximumDotLength;
			result.MinimumWordSpace = this->m_MinimumWordSpace;
			result.TxDashMultiplier = this->m_TxDashMultiplier;
			result.TxWordMultiplier = this->m_TxWordMultiplier;
			return result;
		}
	}
}

#endif
