//
//
//
//		Interfaces.h
//
//		License: GNU General Public License Version 3.0.
//
//		Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY. All rights reserved.
//
//		This program is free software: you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation, either version 3 of the License, or
//		(at your option) any later version.
//
//		This program is distributed in the hope that it will be useful, but
//		WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//		General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program.  If not, see: http://www.gnu.org/licenses/
//
//
//

#ifndef __MODEM_INTERFACES_H
#define __MODEM_INTERFACES_H

#include "Elements.h"
#include <float.h>

using namespace KK5JY::Collections;

namespace KK5JY {
	namespace CW {
		// speed control states
		enum SpeedSources {
			SpeedManual = 0,
			SpeedAuto = 1
		};

		/// <summary>
		/// Translates detected elements into a logical symbol stream.
		/// </summary>
		class TimingLogic {
			public: // dtors
				//
				//  virtual dtor
				//
				virtual ~TimingLogic() { /* nop */ }

			public: // properties
				/// <summary>
				/// Return the current boxcar length.
				/// </summary>
				virtual uint8_t BoxCarLength() const = 0;

				/// <summary>
				/// Set a new boxcar length.
				/// </summary>
				virtual void BoxCarLength (unsigned bcLength) = 0;

				/// <summary>
				/// Return the current average dot length from the tracker.
				/// </summary>
				virtual float DotLength() const = 0;

				/// <summary>
				/// Return the current safety gap.
				/// </summary>
				virtual float SafetyGap() const = 0;

				/// <summary>
				/// Estimate the current RX WPM based on the average dot length.
				/// </summary>
				virtual float RxWPM() const = 0;

				/// <summary>
				/// Estimate the current RX WPM based on the average dot length.
				/// </summary>
				virtual float TxWPM() const = 0;

				/// <summary>
				/// The minimum length for a word space, as a multiple of the current average dot length.
				/// </summary>
				virtual float MinimumWordSpace() const = 0;

				/// <summary>
				/// Set the new minimum word space.
				/// </summary>
				virtual void MinimumWordSpace(float value) = 0;

				/// <summary>
				/// The distance, as a fraction of the average, that a new sample's pulse width must
				/// be in order to be counted in the average.  This prevents the average from collapsing
				/// in on itself when a long string of dots or dashes is encountered.
				/// </summary>
				virtual float MinimumAverageDistance() const = 0;

				/// <summary>
				/// Set the new minimum average distance (see above).
				/// </summary>
				virtual void MinimumAverageDistance(float mad) = 0;

				/// <summary>
				/// Set the RX WPM, and drop into manual RX mode.
				/// </summary>
				virtual void RxWPM(int wpm) = 0;

				/// <summary>
				/// Set the TX WPM, and drop into manual TX mode.
				/// </summary>
				virtual void TxWPM(int wpm) = 0;

				//
				//  Get the RX speed source.
				//
				virtual SpeedSources RxMode() const = 0;

				//
				//  Get the TX speed source.
				//
				virtual SpeedSources TxMode() const = 0;

				//
				//  Set the RX speed source.
				//
				virtual void RxMode(SpeedSources src) = 0;

				//
				//  Set the TX speed source.
				//
				virtual void TxMode(SpeedSources src) = 0;

				//
				//  Get the TX dash multiplier.
				//
				virtual uint8_t DashMultiplier() const = 0;

				//
				//  Set the TX dash multiplier.
				//
				virtual void DashMultiplier(uint8_t value) = 0;

				//
				//  Get the TX word multiplier.
				//
				virtual uint8_t WordMultiplier() const = 0;

				//
				//  Set the TX word multiplier.
				//
				virtual void WordMultiplier(uint8_t value) = 0;

			public: // methods
				/// <summary>
				/// Do the decoding.
				/// </summary>
				/// <param name="raw">The raw element data.</param>
				/// <param name="result">A decoded element symbol stream.</param>
				/// <returns>True if a word space was added to the result, indicating data ready to decode.</returns>
				virtual bool Decode(CircularBuffer<CwElement> &raw, CircularBuffer<DataElement> &result) = 0;

				/// <summary>
				/// Do the decoding.
				/// </summary>
				virtual int Encode(CircularBuffer<DataElement> &txBuffer, CircularBuffer<CwElement> &cwBuffer) = 0;
		};

		class DecoderLogic {
			public: // dtors
				virtual ~DecoderLogic() { /* nop */ }

			public:
				/// <summary>
				/// Do the decoding.
				/// </summary>
				/// <param name="raw">The raw element stream.</param>
				/// <returns>Decoded text.</returns>
				virtual int Decode(CircularBuffer<DataElement> &rxBuffer, char *buffer, int buflen) = 0;

				/// <summary>
				/// Do the encoding.
				/// </summary>
				virtual int Encode(CircularBuffer<DataElement> &txBuffer, const char *buffer, int bufLen) = 0;

				/// <summary>
				/// The character to print if decoding fails for a single character.
				/// </summary>
				virtual char ErrorSymbol() const = 0;

				/// <summary>
				/// Set the error character.
				/// </summary>
				virtual void ErrorSymbol(char value) = 0;
		};
	}
}

#endif
