/*
 *
 *
 *    TriGraph.h
 *
 *    A five-state graph of LEDs used as a centering indicator.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

#ifndef __KK5JY_TRIGRAPH_H
#define __KK5JY_TRIGRAPH_H

#define TRIGRAPH_LENGTH (3)

namespace KK5JY {
	//
	//  TriGraph - drive a dot-graph of LEDs on discrete digital pins
	//
	class TriGraph {
		private:
			// pin array
			uint8_t pins[TRIGRAPH_LENGTH];

		public:
			//
			//  Ranges - these are the ranges for each LED; a typical triagraph would look like
			//           the following, which gives roughly equal range to each of the five states.
			//
			//      +-- (-127)               +-------------+---- InnerBlankingEdge
			//      +                        +             +
			//      [ ******** LEFT ******** |      0      | ********* RIGHT ******** ]   << OUTER LEDS
			//      [             | ************ CENTER ************* |               ]   << CENTER LED
			//                    +                                   +               +
			//                    +-------- OuterBlankingEdge --------+         127 --+
			//

			// this is the value (measured from zero) beyond which one of the outer LEDs is illuminated
			uint8_t InnerBlankingEdge = 23;

			// this is the value (measured from zero) inside which the center LED is illuminated
			uint8_t OuterBlankingEdge = 75;

		public:
			// ctor - assign pins and number of pins
			TriGraph(const uint8_t _left, const uint8_t _center, const uint8_t _right);

			// update the graph; values in range of (-127, 127)
			void update(int8_t newValue) const;

			// clear the graph, turning off all LEDs
			void clear(void) const;

			// run test sequence on LEDs
			void test(uint16_t msec = 100) const;

			// setup
			void setup(void) const;
	};

	//
	//  ctor(...)
	//
	inline TriGraph::TriGraph(const uint8_t _left, const uint8_t _center, const uint8_t _right) {
		pins[0] = _left;
		pins[1] = _center;
		pins[2] = _right;
	}


	//
	//  setup()
	//
	inline void TriGraph::setup() const {
		for (uint8_t i = 0; i != TRIGRAPH_LENGTH; ++i) {
			pinMode(pins[i], OUTPUT);
			digitalWrite(pins[i], LOW);
		}
	}


	//
	//  test()
	//
	inline void TriGraph::test(uint16_t msec) const {
		for (uint8_t i = 0; i != TRIGRAPH_LENGTH; ++i) {
			digitalWrite(pins[i], HIGH);
			delay(msec);
		}
		for (uint8_t i = 0; i != TRIGRAPH_LENGTH; ++i) {
			digitalWrite(pins[i], LOW);
			delay(msec);
		}
	}

	//
	//  update(...)
	//
	inline void TriGraph::update(int8_t newValue) const {
		digitalWrite(pins[0], (newValue < (-InnerBlankingEdge)) ? HIGH : LOW);
		digitalWrite(pins[1], (newValue > (-OuterBlankingEdge) && newValue < (OuterBlankingEdge)) ? HIGH : LOW);
		digitalWrite(pins[2], (newValue > (InnerBlankingEdge)) ? HIGH : LOW);
	}

	//
	//  clear()
	//
	inline void TriGraph::clear(void) const {
		for (uint8_t j = 0; j != TRIGRAPH_LENGTH; ++j) {
			digitalWrite(pins[j], LOW);
		}
	}
}

#endif
