/*
 *
 *
 *    avraudio2.h
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *    Portions adapted from Instructables article by Amanda Ghassaei
 *      http://www.instructables.com/id/Arduino-Audio-Input/
 *      Sept 2012, Also under GPL3.
 *
 *
 */


#ifndef __KK5JY_AVRAUDIO2_H
#define __KK5JY_AVRAUDIO2_H

//
//  ComputeTimer1(...) - compute the Timer2 limit register for a specific interrupt rate
//
inline static uint32_t ComputeTimer1(double f) {
	return round((F_CPU / f) - 1);
}

//
//  StartAudioInput(f) - configure A/D for audio input with sampling rate
//
//     'f' is the sampling rate desired
// 
//     'eightbit' should be true if the user is only reading the top eight bits from
//     the A/D (the ADCH register); value of false should be used if reading the entire
//     10-bit range of the A/D
//
//  Use of this function requires two ISRs:
//
//     ISR(ADC_vec)
//        the interrupt handler called when A/D value ready; value can be
//        read either from ADCH (8-bit mode) or ADC (10-bit mode)
//
//     ISR(TIMER1_COMPB_vect)
//        the interrupt handler called when the Timer1 CTC is reached; this
//        handler can be empty, but it must exist in the sketch, because the
//        A/D trigger depends on the interrupt that generates this handler.
//
inline void StartAudioInput(double f, bool eightbit = false) {
	// disable interrupts
	cli();

	// clear ADC control registers
	ADCSRA = 0;
	ADCSRB = 0;

	// set reference voltage to external AREF pin, this is 5V (or 3.3V) on Arduino
	ADMUX |= (1 << REFS0);

	// if 8-bit mode, left-align the ADC value; ISR should read ADCH; otherwise read ADC register (10 bits)
	if (eightbit)
		ADMUX |= (1 << ADLAR);

	// set ADC clock prescaler to 128; for a 16MHz CPU clock, this gives 125kHz A/D clock
	ADCSRA |= (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
	ADCSRA |= (1 << ADATE);		// enable auto trigger
	ADCSRA |= (1 << ADIE);		// enable interrupts when measurement complete
	ADCSRA |= (1 << ADEN);		// enable ADC
	ADCSRA |= (1 << ADSC);		// start ADC measurements

	// trigger A/D on Timer1 Compare Match B interrupt
	ADCSRB |= (1 << ADTS2) | (1 << ADTS0);

	//
	//  Run Timer1 in CTC mode, with 16-bit divisor
	//
	TCCR1A = 0;					// clear Timer1 control registers
	TCCR1B = 0;
	TCNT1 = 0;					// set Timer1 counter to zero
	TCCR1B |= (1 << WGM12);		// enable CTC mode; limit is OCR1A
	TCCR1B |= (1 << CS10);		// set prescaler to 1x
	TIMSK1 = (1 << OCIE1B);		// Output Compare Interrupt Enable B

	// configure timer limit registers; TODO: some more research is needed here,
	//    but what appears to be happening is that the CTC mode can only use the
	//    OCR1A register as a limit, but it's the OCR1B register that drives the
	//    A/D trigger; so both limit registers have to be set to the same value.
	OCR1A =						// set output compare register A
	OCR1B = ComputeTimer1(f);	// set output compare register B

	// enable interrupts
	sei();
}

//
//  StopAudioInput() - stop the A/D sampling started with StartAudioInput(...)
//
inline void StopAudioInput(void) {
	// disable interrupts
	cli();

	TIMSK1 = 0;					// disable Timer1 interrupts
	ADCSRA &= ~(1 << ADATE);	// disable A/D auto trigger
	ADCSRA &= ~(1 << ADIE);		// disable A/D interrupts
	ADCSRA &= ~(1 << ADEN);		// disable ADC
	ADCSRA &= ~(1 << ADSC);		// stop ADC measurements

	// enable interrupts
	sei();
}

#endif
