/*
 *
 *
 *    config_A0.h
 *
 *    Configuration options for reading CW data from Teensy audio adaptor.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2017 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

//
//  REQUIRED: Set the center frequency (Hz) of the tone tetector.
//
#define CW_FREQ (700.0)

//
//  REQUIRED: Set the bandwidth of the input filter.
//
#define CW_BW (500.0)

//
//  REQUIRED: Set the order of the input filter.
//
//  Zero will disable the filter entirely, valid values are one through four.
//  Higher values will require slightly more CPU time, but will make the
//  transition bands sharper.
//
#define CW_ORDER (1)

//
//  REQUIRED: Set the input volume level; valid values are 0 to 15,
//  and correspond to the following levels on the Teensy audio shield:
//
//   0: 3.12 Volts p-p
//   1: 2.63 Volts p-p
//   2: 2.22 Volts p-p
//   3: 1.87 Volts p-p
//   4: 1.58 Volts p-p
//   5: 1.33 Volts p-p
//   6: 1.11 Volts p-p
//   7: 0.94 Volts p-p
//   8: 0.79 Volts p-p
//   9: 0.67 Volts p-p
//   10: 0.56 Volts p-p
//   11: 0.48 Volts p-p
//   12: 0.40 Volts p-p (recomended default)
//   13: 0.34 Volts p-p
//   14: 0.29 Volts p-p
//   15: 0.24 Volts p-p
//
#define CW_INPUT_LEVEL (12)

//
//  REQUIRED: Tone detection algorithm selection.
//
//  Set this to 'true' to enable LM567 emulation in the firmware.
//  This will still read audio data from the A0 pin, but the
//  detection algorithm will emulate an LM567 chip in software.
//
//  Set this to 'false' to use a delay-line-product algorithm.
//
//  The 567 emulation performance is a few dB better when trying
//  to detect a tone buried in wideband Gaussian noise.
//
#define EMULATE_LM567 true

//
//  ADVANCED: Set the sample rate divisor
//
//  This is the downsampling factor to run after the IIR preselector.  Values
//  between 1 and 8 are appropriate for CW frequencies between 400 and 2000Hz.
//  A value of 4 is a good starting place.
//
//  Different divisors will produce different effective sampling rates for
//  the detector algorithm.  The preselector will run at the codec's 44100Hz,
//  regardless of this setting.
//
//      Divisor		Fs
//         1		44100
//         2		22050
//         3		14700
//         4		11025 (recomended default)
//         5		8820
//         6		7350
//         7		6300
//         8		5512.5
//
//  The preselector will also serve as an antialiasing filter, allowing the
//  detector to run at a lower sampling frequency, which will require less
//  CPU processing power.
//
#define CW_RATE_DIVISOR (4)

//
//  ADVANCED: enable tuning indicator
//
//  If using the 567 emulation, setting these will enable a trigraph
//  of LEDs on the corresponding pins, which can be used as a tuning
//  indicator.
//
#define CW_TRIGRAPH_LEFT (5)
#define CW_TRIGRAPH_CENTER (6)
#define CW_TRIGRAPH_RIGHT (7)

//
//  ADVANCED: these control certain PLL settings when EMULATE_LM567 is
//  set to 'true'.  These have been tuned to have good performance when
//  used with the Teensy audio shield, which uses a 44.1kHz sampling rate.
//
//  Setting any of these value to zero will cause a reasonable default
//  value to be computed at runtime and used.
//
//  Do not change any of these settings unless you know what you are doing.
//  These numbers assume a CW_RATE_DIVISOR = 4.
//
//  These are the defaults used when the program is first run; each of
//  these can be changed at runtime (and the changes stored in EEPROM)
//  using appropriate commands on the serial/host connection.
//
#define EM567_OUTPUT_INTEGRATOR 350
#define EM567_LOOP_INTEGRATOR 44
#define EM567_LOOP_VCO_LIMIT 1
//
//  NOTE: Some reasonable bench-tested settings for different divisors are below:
//
//  CW_RATE_DIVISOR = 1 (fs = 22050):
//      EM567_OUTPUT_INTEGRATOR 1000
//      EM567_LOOP_INTEGRATOR 40
//      EM567_LOOP_VCO_LIMIT 5
//
//  CW_RATE_DIVISOR = 3 (fs = 14700):
//		EM567_OUTPUT_INTEGRATOR 306
//		EM567_LOOP_INTEGRATOR 35
//		EM567_LOOP_VCO_LIMIT 1
//
//  CW_RATE_DIVISOR = 4 (fs = 11025):
//      EM567_OUTPUT_INTEGRATOR 230
//      EM567_LOOP_INTEGRATOR 26
//      EM567_LOOP_VCO_LIMIT 1
//
//  CW_RATE_DIVISOR = 6 (fs = 7350):
//      EM567_OUTPUT_INTEGRATOR 100
//      EM567_LOOP_INTEGRATOR 20
//      EM567_LOOP_VCO_LIMIT 1
//

//
//  ADVANCED: The control voltage LPF output offset
//
//  These do not affect the detection algorithm; they only control the
//  analysis of the PLL control voltage, which can be used for tuning
//  indicators, and the #DEV; command, which does the same thing through
//  a host command.
//
#define EM567_TRACK_OFFSET (-234)
#define EM567_TRACK_GAIN (350)
#define EM567_TRACK_ALPHA (0.00035)
#define EM567_TRACK_LOW_SCALE (0.72)

//
//  ADVANCED: PLL tracking unlock delay (msec)
//
//  This controls how long after the PLL unlocks that the last measured
//  PLL offset is reported to the #DEV; command and the LED trigraph, if
//  configured.  The firmware only tracks the PLL offset when the PLLL is
//  locked, but when the PLL unlocks, the last locked measurement remains
//  latched and available for this many msec before the offset is reported
//  as zero, and the trigraph extinguished.
//
#define PLL_OFFSET_LATCH_DELAY (500)

// EOF:
