/*
 *
 *
 *    EEBlock.h
 *
 *    EEPROM management library
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef __EEBLOCK_H
#define __EEBLOCK_H

#include <Arduino.h>
#include <EEPROM.h>
#include <stdlib.h>
#include <string.h>

//
//  EEPROM data block with write-levelling; this class can be instantiated
//      as a const instance, since the I/O methods don't modify the data.
//
class EEBlock {
	private:
		const char *magic;
		const int size;
		
	public:
		EEBlock(const char *ms, int sz = 0) : magic(ms), size(sz == 0 ? (E2END + 1) : sz) { /* nop */ }

	public:
		// read the existing EEPROM block; returns number of chars read, or (-1) on failure
		int read(char *buffer, int count) const;
		
		// write a new EEPROM block; the block will be moved, if possible
		int write(char *buffer, int count) const;

		// finds the block within the EEPROM, and returns the offset, or (-1) on failure
		int seek() const;

		// clear the entire EEPROM
		void clear() const;
};

//
//  ReadLength(offset) - deserialize the length encoded at a specific EEPROM address
//
static int ReadLength(int offset) {
	int len = 0;
	*((char*)(&len)) = EEPROM.read(offset++);
	*(((char*)(&len)) + 1) = EEPROM.read(offset);
	return len;
}

//
//  WriteLength(offset, length) - serialize the length at a specific EEPROM address
//
static void WriteLength(int offset, int len) {
	EEPROM.write(offset++, *((char*)(&len)));
	EEPROM.write(offset, *(((char*)(&len)) + 1));
}

//
//  clear() - erase the EEPROM
//
inline void EEBlock::clear() const {
	for (int i = 0; i != size; ++i) {
		EEPROM.write(i, 0);
	}
}

//
//  read() - get the existing EEPROM block; returns number of chars read, or (-1) on failure
//
inline int EEBlock::read(char *buffer, int count) const {
	int offset = seek();
	if (offset < 0) return -1;

	// read the length field
	int pos = offset + strlen(magic);
	int len = ReadLength(pos);
	pos += 2;

	// don't overrun the buffer
	if (len > count) {
		len = count;
	}

	// read the data
	for (int i = 0; i != len; ++i) {
		*buffer++ = EEPROM.read(pos++);
	}

	// return the number of bytes read
	return len;
}

//
//  write() - commit a new EEPROM block; the block will be moved, if possible
//
inline int EEBlock::write(char *buffer, int count) const {
	const int mslen = strlen(magic);	// measure the input string
	int offset = seek();				// look for an existing block
	// if no block exists, start at offset zero
	if (offset < 0) {
		offset = 0;
	} else {
		// otherwise, determine where to drop the new block
		int len = ReadLength(offset + mslen);
		int newOffset = offset + mslen + len + 2; // start + magic + data + length
		if (newOffset + mslen + 2 + count > size) {
			newOffset = 0;
		}
		
		// if moving the block, zero out the old block
		if (offset != newOffset) {
			for (int i = 0; i != mslen + len + 2; ++i) {
				EEPROM.write(offset + i, 0);
			}
		}
		
		// store the new location
		offset = newOffset;
	}

	// write the new magic string preamble
	for (int i = 0; i != mslen; ++i) {
		EEPROM.write(offset + i, magic[i]);
	}
	// write out the length field
	WriteLength(offset + mslen, count);
	// write out the payload
	for (int i = 0; i != count; ++i) {
		EEPROM.write(offset + mslen + 2 + i, buffer[i]);
	}

	// return the number of payload bytes written
	return count;
}

inline int EEBlock::seek() const {
	const int mslen = strlen(magic);	// measure the magic string
	char buffer[mslen + 1];				// create an I/O buffer for reading the EEPROM
	
	// build the initial test string
	for (int i = 0; i != mslen; ++i) {
		buffer[i] = EEPROM.read(i);
	}
	buffer[mslen] = 0;

	// now step through the EEPROM, shifting the test string each time
	int pos = 0;
	while (pos < (size - mslen)) {
		#ifdef REALLY_REALLY_VERBOSE_DEBUG
		Serial.print("DEBUG: compare '");
		Serial.print(buffer);
		Serial.print("' to '");
		Serial.print(magic);
		Serial.println("'");
		#endif

		// if the test string matches at the current position, return that position
		if (strcmp(buffer, magic) == 0) {
			return pos;
		}
		// otherwise, shift the test string and increment the position counter
		for(int i = 0; i != (mslen - 1); ++i) {
			buffer[i] = buffer[i + 1];
		}
		buffer[mslen - 1] = EEPROM.read(pos++ + mslen);
	}

	// return failure
	return -1;
}

#endif
