/*
 *
 *
 *    em567.h
 *
 *    Tone detector, emulating the LM567 circuit.
 *
 *    Copyright (C) 2017 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_EM567_H
#define __KK5JY_EM567_H

#include "cavg.h"
#include "vco.h"
#include "sign.h"
#include "itone.h"

//#define VERBOSE_DEBUG
#ifdef VERBOSE_DEBUG
#include <sstream>
#endif

// full-scale in the integrator
#define INTEGRATOR_FULL_SCALE 20

// hysteresis limits (range is zero to INTEGRATOR_FULL_SCALE)
#define HYS_HIGH 12
#define HYS_LOW 8

// default integrator lengths
#define DEFAULT_INT_I 96
#define DEFAULT_INT_Q 32

// default maximum control voltage
#define DEFAULT_CV_MAX 1

namespace KK5JY {
	namespace DSP {
		//
		//  Tone567 - class for detecting specific tone in audio stream
		//
		template <typename sample_t = double, typename vco_counter_t = uint8_t>
		class Tone567 : public IToneDetector<sample_t> { 
			public:
				struct Config {
					// center frequency
					double f0;

					// sampling frequency
					size_t fs;

					// VCO integrator lengths
					size_t vco_int_I;
					size_t vco_int_Q;

					// VCO max control voltage
					int8_t vco_cv_max;

					// ctors
					Config();
					Config(const Config &c);

					// factory
					static Config Create(double _f0, size_t _fs, size_t vco_i = 0, size_t vco_q = 0, int8_t max_cv = 0);
				};

			private:
				// the VCO and its control voltage
				VCO<sample_t, vco_counter_t> vco;
				int8_t vco_cv;
				Averager<sample_t> vco_I, vco_Q;

				// output state tracking
				bool state;

			public:
				Tone567(const Config &c);

			public:
				// run one sample of the detector algorithm
				bool run(sample_t sample);

				sample_t I() const { return vco_I.value(); }
				sample_t Q() const { return vco_Q.value(); }
				sample_t CV() const { return vco_cv; }

				size_t GetLengthI() const { return vco_I.size(); }
				size_t GetLengthQ() const { return vco_Q.size(); }
				int8_t GetMaxCV() const { return vco.rail(); }

				void SetLengthI(size_t n) { vco_I.resize(n); }
				void SetLengthQ(size_t n) { vco_Q.resize(n); }
				void SetMaxCV(uint8_t n) { vco.rail(n); }

				sample_t Period() const { return vco.period(); }
				sample_t EffectivePeriod() const { return vco.eff_period(); }
		};

		//
		//  Config ctor
		//
		template <typename sample_t, typename vco_counter_t>
		inline Tone567<sample_t, vco_counter_t>::Config::Config() {
			f0 = 1000;
			fs = 8000;
			vco_int_I = DEFAULT_INT_I;
			vco_int_Q = DEFAULT_INT_Q;
			vco_cv_max = DEFAULT_CV_MAX;
		}

		//
		//  Config ctor
		//
		template <typename sample_t, typename vco_counter_t>
		inline Tone567<sample_t, vco_counter_t>::Config::Config(const Config &c) {
			f0 = c.f0;
			fs = c.fs;
			vco_int_I = c.vco_int_I;
			vco_int_Q = c.vco_int_Q;
			vco_cv_max = c.vco_cv_max;
		}

		template <typename sample_t, typename vco_counter_t>
		typename Tone567<sample_t, vco_counter_t>::Config
			Tone567<sample_t, vco_counter_t>::Config::Create(double _f0, size_t _fs, size_t _vco_int_I, size_t _vco_int_Q, int8_t _vco_cv_max) {
			Tone567<sample_t, vco_counter_t>::Config result;
			result.f0 = _f0;
			result.fs = _fs;
			if (_vco_int_I == 0) {
				_vco_int_I = DEFAULT_INT_I * _fs / 8000;
			}
			if (_vco_int_Q == 0) {
				_vco_int_Q = DEFAULT_INT_Q * _fs / 8000;
			}
			if (_vco_cv_max == 0) {
				_vco_cv_max = DEFAULT_CV_MAX * _fs / 8000;
			}
			result.vco_int_I = _vco_int_I;
			result.vco_int_Q = _vco_int_Q;
			result.vco_cv_max = _vco_cv_max;
			return result;
		}

		//
		//  Tone567 ctor
		//
		template <typename sample_t, typename vco_counter_t>
		inline Tone567<sample_t, vco_counter_t>::Tone567(const Config &c)
			: vco(c.f0, c.fs, c.vco_cv_max),
			  vco_cv(0),
			  vco_I(c.vco_int_I),
			  vco_Q(c.vco_int_Q),
			  state(false) {
			// nop
		}

		//
		//  Tone567::run - this is the detector algorithm
		//
		template <typename sample_t, typename vco_counter_t>
		inline bool Tone567<sample_t, vco_counter_t>::run(sample_t sample) {
			// run the VCO and the phase detectors
			sample = sign(sample);
			VCO_SampleIQ<sample_t> vout = vco.run(vco_cv);
			VCO_SampleIQ<sample_t> samples(
				INTEGRATOR_FULL_SCALE * vout.I * sample,
				INTEGRATOR_FULL_SCALE * vout.Q * sample
			);
			// phase detector integrators
			vco_cv = vco_Q.run(samples.Q);
			vco_I.add(samples.I);

			// return the current state
			if (state) {
				state = vco_I.value() > HYS_LOW;
			} else {
				state = vco_I.value() > HYS_HIGH;
			}
			return state;
		}
	}
}

#endif
