/*
 *
 *
 *    es.h
 *
 *    Exponential smoothers.
 *
 *    Copyright (C) 2017 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY__EXPS
#define __KK5JY__EXPS

#include <stdint.h>
#include "nlimits.h"

namespace KK5JY {
	namespace DSP {
		// ---------------------------------------------------------------------------
		//
		//   class Smoother<T>
		//
		//   Exponential smoothing class for floating point types.
		//
		// ---------------------------------------------------------------------------
		template <typename T>
		class Smoother {
			public:
				const T Alpha;
			private:
				T state;
			public:
				Smoother(T alpha, T seed = 0);
				T run(T in);
				T value() const { return state; }
		};

		//
		//  Smoother::ctor
		//
		template <typename T>
		Smoother<T>::Smoother(T alpha, T seed) : Alpha(alpha), state(seed) { }

		//
		//  Smoother::run
		//
		template <typename T>
		T Smoother<T>::run(T in) {
			return state = (in * Alpha) + (state * (1 - Alpha));
		}

		//
		//  Calculate alpha value to emulate RC lowpass filter in smoother.
		//
		double LowpassToAlpha(size_t fs, double fc) {
			// caluclate the equivalent R-C constant
			double rc = 1.0 / (2.0 * M_PI * fc);

			// calculate the sampling period
			double dt = 1.0 / fs;

			// calculate equivalent alpha value
			return dt / (rc + dt);
		}

		//
		//  calculate lowpass fc value from alpha when using smoother as LP filter.
		//
		double AlphaToLowpass(double alpha, size_t fs) {
			// calculate the sampling period
			double dt = 1.0 / fs;

			// calculate RC value from alpha
			double rc = (dt * (1.0 - alpha)) / alpha;

			// caluclate fc from RC
			return 1.0 / (2.0 * M_PI * rc);
		}

		// ---------------------------------------------------------------------------
		//
		//   class FixedSmoother<T>
		//
		//   Exponential smoothing class for fixed point types.
		//
		// ---------------------------------------------------------------------------
		template <typename T>
		class FixedSmoother {
			public:
				const double Alpha;
			private:
				T fixedAlpha;
				T state;
			public:
				FixedSmoother(double alpha, T seed = 0);
				T run(T in);
				T value() const { return state; }
		};

		//
		//  FixedSmoother::ctor
		//
		template <typename T>
		FixedSmoother<T>::FixedSmoother(double alpha, T seed)
			: Alpha(alpha),
			  fixedAlpha(alpha * KK5JY::DSP::norm_limits<T>::maximum),
			  state(seed) {
			// nop
		}

		//
		//  FixedSmoother<T>::run(...) -- this is the generic version, for use with unsigned math only
		//
		template <typename T>
		T FixedSmoother<T>::run(T in) {
			return state =
				(((in * fixedAlpha) +
				((state * (KK5JY::DSP::norm_limits<T>::maximum - fixedAlpha))))
					>> ((sizeof(T) * 8))); // i.o.w., / KK5JY::DSP::norm_limits<T>::maximum);
		}

		//
		//  FixedSmoother<uint8_t>::run(...)
		//
		template <>
		uint8_t FixedSmoother<uint8_t>::run(uint8_t in) {
			return state =
				(((static_cast<uint16_t>(in) * fixedAlpha) +
				((static_cast<uint16_t>(state) * (static_cast<uint16_t>(KK5JY::DSP::norm_limits<uint16_t>::maximum) - fixedAlpha))))
					>> ((sizeof(uint8_t) * 8))); // i.o.w., / KK5JY::DSP::norm_limits<uint8_t>::maximum);
		}

		//
		//  FixedSmoother<int8_t>::run(...)
		//
		template <>
		int8_t FixedSmoother<int8_t>::run(int8_t in) {
			return state =
				(((static_cast<int16_t>(in) * fixedAlpha) +
				((static_cast<int16_t>(state) * (static_cast<int16_t>(KK5JY::DSP::norm_limits<int8_t>::maximum) - fixedAlpha))))
					>> ((sizeof(int8_t) * 8) - 1)); // i.o.w., / KK5JY::DSP::norm_limits<int16_t>::maximum);
		}

		//
		//  FixedSmoother<uint16_t>::run(...)
		//
		template <>
		uint16_t FixedSmoother<uint16_t>::run(uint16_t in) {
			return state =
				(((static_cast<uint32_t>(in) * fixedAlpha) +
				((static_cast<uint32_t>(state) * (static_cast<uint32_t>(KK5JY::DSP::norm_limits<uint32_t>::maximum) - fixedAlpha))))
					>> ((sizeof(uint16_t) * 8))); // i.o.w., / KK5JY::DSP::norm_limits<uint16_t>::maximum);
		}

		//
		//  FixedSmoother<int16_t>::run(...)
		//
		template <>
		int16_t FixedSmoother<int16_t>::run(int16_t in) {
			return state =
				(((static_cast<int32_t>(in) * fixedAlpha) +
				((static_cast<int32_t>(state) * (static_cast<int32_t>(KK5JY::DSP::norm_limits<int16_t>::maximum) - fixedAlpha))))
					>> ((sizeof(int16_t) * 8) - 1)); // i.o.w., / KK5JY::DSP::norm_limits<int16_t>::maximum);
		}

		//
		//  FixedSmoother<uint32_t>::run(...)
		//
		template <>
		uint32_t FixedSmoother<uint32_t>::run(uint32_t in) {
			return state =
				(((static_cast<uint64_t>(in) * fixedAlpha) +
				((static_cast<uint64_t>(state) * (static_cast<uint64_t>(KK5JY::DSP::norm_limits<uint32_t>::maximum) - fixedAlpha))))
					>> ((sizeof(uint32_t) * 8))); // i.o.w., / KK5JY::DSP::norm_limits<uint32_t>::maximum);
		}

		//
		//  FixedSmoother<int32_t>::run(...)
		//
		template <>
		int32_t FixedSmoother<int32_t>::run(int32_t in) {
			return state =
				(((static_cast<int64_t>(in) * fixedAlpha) +
				((static_cast<int64_t>(state) * (static_cast<int64_t>(KK5JY::DSP::norm_limits<int32_t>::maximum) - fixedAlpha))))
					>> ((sizeof(int32_t) * 8) - 1)); // i.o.w., / KK5JY::DSP::norm_limits<int32_t>::maximum);
		}
	}
}

#endif // __KK5JY__EXPS
