/*
 *
 *
 *    tone.h
 *
 *    Delay line-based Tone detector.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_TONE_H
#define __KK5JY_TONE_H

#include "sign.h"
#include "cavg.h"
#include "minmax.h"
#include "nlimits.h"
#include "itone.h"
#ifdef __linux__
#include "ops.h"
#endif

//#define VERBOSE_DEBUG
#ifdef VERBOSE_DEBUG
#include <sstream>
#endif

// define hysteresis loop rails (normalized)
#define TONE_HYS_LOW (0.40)
#define TONE_HYS_HIGH (0.60)

namespace KK5JY {
	namespace DSP {
		//
		//  ToneDetector - class for detecting specific tone in audio stream
		//
		template <typename sample_t = double, typename integrator_t = Averager<sample_t>, typename interpolator_t = double >
		class ToneDetector : public IToneDetector<sample_t> { 
			public:
				struct Config {
					// center frequency
					double f0;
					// sampling frequency
					size_t fs;
					// STA window (wavelengths)
					size_t sta_win;
					// LTA window (wavelengths)
					size_t lta_win;
					// minimum envelope magnitude (S-N; normalized)
					double th_min; // TODO: make the minimum threshold a specific minimum dB value (e.g., 1 bit = 3dB)

					// ctors
					Config();
					Config(const Config &c);

					// factory
					static Config Create(double _f0, size_t _fs, size_t _sta = 0, size_t _lta = 0, double _min = 0);
				};

			public:
				// 'N' is the delay line length
				const double N;
				// 'dist' is the whole part of 'N'
				const size_t dist;
				// 'frac' is the fractional part of 'N'
				const interpolator_t frac;
				// 'frac_1m' is equal to 1.0 - frac
				const interpolator_t frac_1m;
				// 'th_min' is the minimum envelope magnitude (S-N)
				const sample_t th_min;

			private:
				// divide by 2
				const static interpolator_t divBy2;
				// length of the delay line
				const size_t length;
				// pointer to the head of the delay line
				sample_t * const delay;
				// input pointer
				sample_t *ptr;

				// min-max tracking for threshold formation
				MinMaxTrack<sample_t> minmax;
				// envelope integrator - forms an short-term average of the absolute value of input
				integrator_t sta;
				// signal integrator - LTA of envelope max
				integrator_t signal;
				// noise floor integrator - LTA of envelope min, essentially the average noise mangitude
				integrator_t noise;
				// thresholds (one on each end of the hysteresis loop)
				sample_t th_low, th_high;
				// output state tracking
				bool state;

			public:
				ToneDetector(const Config &c);

			public:
				// run one sample of the detector algorithm
				bool run(sample_t sample);

				// return the current envelope level
				int32_t level() const { return sta.value(); }
		};

		//
		//  Config ctor
		//
		template <typename sample_t, typename integrator_t, typename interpolator_t>
		inline ToneDetector<sample_t, integrator_t, interpolator_t>::Config::Config() {
			f0 = 1000;
			fs = 8000;
			sta_win = 5;
			lta_win = 50;
			th_min = 0.05;
		}

		//
		//  Config ctor
		//
		template <typename sample_t, typename integrator_t, typename interpolator_t>
		inline ToneDetector<sample_t, integrator_t, interpolator_t>::Config::Config(const Config &c) {
			f0 = c.f0;
			fs = c.fs;
			sta_win = c.sta_win;
			lta_win = c.lta_win;
			th_min = c.th_min;
		}

		template <typename sample_t, typename integrator_t, typename interpolator_t>
		typename ToneDetector<sample_t, integrator_t, interpolator_t>::Config
			ToneDetector<sample_t, integrator_t, interpolator_t>::Config::Create(
				double _f0, size_t _fs, size_t _sta, size_t _lta, double _min) {
			ToneDetector<sample_t, integrator_t, interpolator_t>::Config result;
			result.f0 = _f0;
			result.fs = _fs;
			if (_sta != 0)
				result.sta_win = _sta;
			if (_lta != 0)
				result.lta_win = _lta;
			if (_min != 0)
				result.th_min = _min;
			return result;
		}

		//
		//  divBy2
		//
		template <typename sample_t, typename integrator_t, typename interpolator_t>
		const interpolator_t ToneDetector<sample_t, integrator_t, interpolator_t>::divBy2(0.5);

		//
		//  ToneDetector ctor
		//
		template <typename sample_t, typename integrator_t, typename interpolator_t>
		inline ToneDetector<sample_t, integrator_t, interpolator_t>::ToneDetector(const Config &c)
			: N(static_cast<double>(c.fs) / c.f0),	// delay line length (exact) = wavelength of f0, in samples
			  dist(static_cast<size_t>(N)),			// delay line length (truncated integer)
			  frac(N - static_cast<size_t>(N)),		// N - dist
			  frac_1m(1.0 - frac),					// 1 - frac
			  th_min(2.0 * c.th_min * norm_limits<sample_t>::maximum), // initialize threshold centerline
			  length(static_cast<size_t>(N) + 2),	// length of delay array = dist + 2 to allow for interpolation
			  delay(new sample_t[length]),			// delay array
			  ptr(delay),							// pointer for walking delay array
			  minmax(c.sta_win * dist),				// min/max tracker
			  sta(c.sta_win * dist),				// short-term average of product envelope
			  signal(c.lta_win / dist),				// LTA of envelope max (technically, this is actually S+N)
			  noise(c.lta_win / dist),				// LTA of envelope min (should be close to the noise level)
			  th_low(th_min * TONE_HYS_LOW),		// initialize threshold low hysteresis line
			  th_high(th_min * TONE_HYS_HIGH),		// initialize threshold high hysteresis line
			  state(false) {						// initialize output state
			// zero out the history elements
			for (size_t i = 0; i != length; ++i) {
				*ptr++ = static_cast<sample_t>(0);
			}
			// initialize the input pointer
			ptr = delay;
		}

		//
		//  ToneDetector::run - this is the detector algorithm
		//
		template <typename sample_t, typename integrator_t, typename interpolator_t>
		inline bool ToneDetector<sample_t, integrator_t, interpolator_t>::run(sample_t sample) {
			// delay line: fetch the sample one wavelength ago ('N' sample delay)
			sample_t *q = ptr + dist;
			if (q >= (delay + length)) {
				q -= length;
			}
			sample_t q1 = *q;

			// interpolate if needed, to get closer to the exact value of 'N'
			if (frac != 0) {
				q = ptr + dist + 1;
				if (q >= (delay + length)) {
					q -= length;
				}
				sample_t q2 = *q;
				q1 = (q2 * frac) + (q1 * frac_1m);
			}

			// compute the product of input and one wavelength ago, and integrate the envelope
			sample_t product = sample * q1;
			sta.add(product);

			// move the delay line pointer backwards through the delay line, and store the current sample
			*ptr = sample;
			if (ptr == delay) {
				ptr = delay + length - 1;
			} else {
				--ptr;
			}

			#ifdef __linux__
			// run noise floor and threshold calculations (linux version)
			if (minmax.run(Operations<sample_t>::abs(product))) {
			#else
			// run noise floor and threshold calculations (arduino version)
			if (minmax.run(abs(product))) {
			#endif
				// first, update the envelope edge integrators
				signal.add(minmax.maximum());
				noise.add(minmax.minimum());

				// compute the thresholded value (actually 2x the threshold, but we'll divide it out later)
				sample_t th_num = (signal.value() + noise.value()) * divBy2;

				// and increase it, if needed
				if (th_num < th_min) {
					th_num = th_min;
				}

				// compute updated thresholds
				th_low = (TONE_HYS_LOW * th_num);
				th_high = (TONE_HYS_HIGH * th_num);

#ifdef VERBOSE_DEBUG
				std::ostringstream outstream;
				outstream
					<< "min(N) = " << minmax.minimum()
					<< "; max(S) = " << minmax.maximum()
					<< "; th_num = " << th_num
					<< "; high = " << th_high
					<< "; low = " << th_low
					<< "; STA(v) = " << sta.value()
					<< "; STA(a) = " << sta.accum();
#ifdef __linux__
				std::cerr << outstream.str() << std::endl;
#else // arduino
				Serial.println(outstream.str());
#endif
#endif
			}

			// run hysteresis on the output to compute the reconstructed binary stream
			if (state) {
				if (sta.value() < th_low)
					state = false;
			} else {
				if (sta.value() > th_high)
					state = true;
			}

			// return the current state
			return state;
		}
	}
}

#endif
