/*
 *
 *
 *    vco.h
 *
 *    A simple square-wave VCO, with both I and Q outputs.
 *
 *    Copyright (C) 2017 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */

#ifndef __KK5JY_VCO_H
#define __KK5JY_VCO_H

#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include "ops.h"

//#define VERBOSE_DEBUG
#ifdef VERBOSE_DEBUG
#include <sstream>
#endif

namespace KK5JY {
	namespace DSP {
		//
		//  A special type for I/Q VCO output
		//
		template <typename sample_t>
		struct VCO_SampleIQ {
			sample_t I; // the in-phase value
			sample_t Q; // the quadrature value

			VCO_SampleIQ() : I(0), Q(0) { }
			VCO_SampleIQ(sample_t i, sample_t q) { I = i; Q = q; }
		};

		//
		//  VCO - class for detecting specific tone in audio stream
		//
		template <
				typename sample_t = double,		// the output sample type
				typename counter_t = uint8_t,	// the internal counter type
				typename cv_t = int8_t>			// the 'control voltage' type; should be signed version of counter_t
		class VCO { 
			private:
				const counter_t m_Period;	// the period counter
				counter_t m_EffPeriod;		// the effective period (m_Period +/- the CV)
				int8_t m_Rail;				// the rail
				counter_t q_ctr;			// the counter

			public:
				VCO(double _f0, size_t _fs, int8_t cv_rail);

			public:
				// run one sample
				VCO_SampleIQ<sample_t> run(cv_t cv);

				// fetch the rail value
				int8_t rail() const { return m_Rail; }

				// set the rail value
				void rail(int8_t r) { m_Rail = r; }

				// fetch the period value
				counter_t period() const { return m_Period; }

				// fetch the current effective period
				counter_t eff_period() const { return m_EffPeriod; }
		};

		//
		//  VCO ctor
		//
		template <typename sample_t, typename counter_t, typename cv_t>
		inline VCO<sample_t, counter_t, cv_t>::VCO(double _f0, size_t _fs, int8_t cv_rail)
			: m_Period(Operations<counter_t>::closest(_fs / _f0)),
			  m_Rail(abs(cv_rail)),
			  q_ctr(0) {
			// nop
		}

		//
		//  VCO::run - this is the detector algorithm
		//
		template <typename sample_t, typename counter_t, typename cv_t>
		inline VCO_SampleIQ<sample_t> VCO<sample_t, counter_t, cv_t>::run(cv_t cv) {
			// clamp max deviation to +/- 'm_Rail' samples
			if (cv < -m_Rail)
				cv = -m_Rail;
			else if (cv > m_Rail)
				cv = m_Rail;

			// calculate effective period
			m_EffPeriod = m_Period + cv;

			// calculate the period /2 and /4
			counter_t p2 = Operations<counter_t>::div2(m_EffPeriod);	// period / 2
			counter_t p4 = Operations<counter_t>::div2(p2);				// period / 4

			// run the Q counter
			++q_ctr;
			if (q_ctr >= m_EffPeriod) {
				q_ctr -= m_EffPeriod;
			}

			// calculate the I counter value from the Q counter
			counter_t i_ctr = q_ctr + p4;
			if (i_ctr >= m_EffPeriod) {
				i_ctr -= m_EffPeriod;
			}

			// determine the phase detector outputs
			sample_t i = (i_ctr >= p2) ? 1 : -1;
			sample_t q = (q_ctr >= p2) ? 1 : -1;

			// return the I/Q pair
			return VCO_SampleIQ<sample_t>(i, q);
		}
	}
}

#endif
